from typing import Literal, Sequence, override
from types import ModuleType
from functools import partial

import numpy as np
import math

from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _ArrayLikeInt, _Scalar, _ArrayLike, _NumPyArray

from lucid._backend.core import (
    fallback,
    operation,
    func_op,
    binary_func_op,
    unary_func_op,
    poly_func_op,
    _FuncOpReturnType,
    _GradFuncType,
)
from lucid._backend.metal import mx


class reshape(operation):
    def __init__(self, shape: _ShapeLike) -> None:
        super().__init__()
        self.shape = shape

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.reshape(self.shape))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.reshape(a.data, self.shape))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return self.result.grad.reshape(*a.shape)


class _reshape_immediate(operation):
    def __init__(self, shape: _ShapeLike) -> None:
        super().__init__()
        self.shape = shape

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.reshape(self.shape))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.reshape(a.data, self.shape))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return self.result.grad.reshape(a.shape)


class squeeze(operation):
    def __init__(self, axis: _ShapeLike | None = None) -> None:
        super().__init__()
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.squeeze(axis=self.axis))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.squeeze(a.data, axis=self.axis))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return self.result.grad.reshape(a.shape)


class unsqueeze(operation):
    def __init__(self, axis: _ShapeLike) -> None:
        super().__init__()
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.expand_dims(a.data, axis=self.axis))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.expand_dims(a.data, axis=self.axis))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.squeeze(self.result.grad, axis=self.axis)


class expand_dims(unsqueeze): ...


class ravel(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.ravel())
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.reshape(a.data, (-1,)))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return self.result.grad.reshape(a.shape)


class stack(operation):
    def __init__(self, axis: int) -> None:
        super().__init__()
        self.axis = axis

    @poly_func_op()
    def cpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data for a in arr]
        self.result = Tensor(np.stack(data_arr, axis=self.axis))

        return self.result, partial(self.__grad__, arr=arr, lib_=np)

    @poly_func_op(device="gpu")
    def gpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [t.data for t in arr]
        self.result = Tensor(mx.stack(data_arr, axis=self.axis))

        return self.result, partial(self.__grad__, arr=arr, lib_=mx)

    def __grad__(self, arr: tuple[Tensor], lib_: ModuleType) -> _GradFuncType:
        split_grads = lib_.split(self.result.grad, len(arr), axis=self.axis)
        return tuple(split_grads)


class hstack(stack):
    def __init__(self) -> None:
        super().__init__(axis=1)

    @override
    @poly_func_op()
    def cpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data if a.ndim > 1 else a.data.reshape(-1, 1) for a in arr]
        self.result = Tensor(np.hstack(data_arr))

        return self.result, partial(self.__grad__, arr=arr, lib_=np)

    @override
    @poly_func_op(device="gpu")
    def gpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [t.data if t.ndim > 1 else t.data.reshape(-1, 1) for t in arr]
        self.result = Tensor(mx.concatenate(data_arr, axis=1))

        return self.result, partial(self.__grad__, arr=arr, lib_=mx)


class vstack(stack):
    def __init__(self) -> None:
        super().__init__(axis=0)

    @override
    @poly_func_op()
    def cpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data if a.ndim > 1 else a.data.reshape(1, -1) for a in arr]
        self.result = Tensor(np.vstack(data_arr))

        return self.result, partial(self.__grad__, arr=arr, lib_=np)

    @override
    @poly_func_op(device="gpu")
    def gpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [t.data if t.ndim > 1 else t.data.reshape(1, -1) for t in arr]
        self.result = Tensor(mx.concatenate(data_arr, axis=0))

        return self.result, partial(self.__grad__, arr=arr, lib_=mx)


class concatenate(operation):
    def __init__(self, axis: int) -> None:
        super().__init__()
        self.axis = axis

    @poly_func_op()
    def cpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data for a in arr]
        self.result = Tensor(np.concatenate(data_arr, axis=self.axis))
        return self.result, partial(self.__grad__, arr=arr)

    @poly_func_op(device="gpu")
    def gpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data for a in arr]
        self.result = Tensor(mx.concatenate(data_arr, axis=self.axis))
        return self.result, partial(self.__grad__, arr=arr)

    def __grad__(self, arr: tuple[Tensor, ...]) -> tuple:
        split_sizes = [a.shape[self.axis] for a in arr]
        grad = self.result.grad
        outputs = []
        start = 0

        for size in split_sizes:
            end = start + size
            slicer = [slice(None)] * grad.ndim
            slicer[self.axis] = slice(start, end)

            outputs.append(grad[tuple(slicer)])
            start = end

        return tuple(outputs)


class pad(operation):
    def __init__(self, pad_width: _ArrayLikeInt, ndim: int) -> None:
        super().__init__()
        self.pad_width = pad_width
        self.pad_with_norm = self._normalize_pad_width(pad_width, ndim)

    def _normalize_pad_width(
        self, pad_width: _ArrayLikeInt, ndim: int
    ) -> _ArrayLikeInt:
        if isinstance(pad_width, int):
            return ((pad_width, pad_width),) * ndim

        if isinstance(pad_width, (tuple, list)):
            pad_width = list(pad_width)
            if all(isinstance(pw, int) for pw in pad_width):
                if len(pad_width) == 1:
                    return ((pad_width[0], pad_width[0]),) * ndim
                elif len(pad_width) == 2:
                    return (tuple(pad_width),) * ndim
                elif len(pad_width) == ndim:
                    return tuple((pw, pw) for pw in pad_width)

            elif all(
                isinstance(pw, (tuple, list)) and len(pw) == 2 for pw in pad_width
            ):
                if len(pad_width) == ndim:
                    return tuple(tuple(pw) for pw in pad_width)
                elif len(pad_width) == 1:
                    return (tuple(pad_width[0]),) * ndim

        raise ValueError(f"Invalid pad_width format: '{pad_width}'.")

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.pad(a.data, self.pad_width))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.pad(a.data, self.pad_width))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad_input = lib_.zeros_like(a.data)
        slices = []
        for pw in self.pad_with_norm:
            before, after = pw
            start = before
            end = -after if after != 0 else None
            slices.append(slice(start, end))

        grad_input = self.result.grad[tuple(slices)]
        return grad_input


class repeat(operation):
    def __init__(self, repeats: int | Sequence[int], axis: int | None) -> None:
        super().__init__()
        self.repeats = repeats
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.repeat(a.data, self.repeats, axis=self.axis))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.repeat(a.data, self.repeats, axis=self.axis))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad_input = lib_.zeros_like(a.data)
        repeats_arr = lib_.array(self.repeats)

        if self.axis is None:
            input_flat = a.data.flatten()
            grad_input_flat = grad_input.flatten()
            grad_output_flat = self.result.grad.flatten()

            input_size = input_flat.size

            if repeats_arr.size == 1:
                repeats_arr = lib_.full(input_size, repeats_arr)
            elif repeats_arr.size != input_size:
                raise ValueError(
                    "repeats must be an integer or a "
                    + "sequence of the same length as input."
                )

            input_indices = lib_.arange(input_size)
            if lib_ is np:
                output_indices = np.repeat(input_indices, repeats_arr)
                np.add.at(grad_input_flat, output_indices, grad_output_flat)
            else:
                output_indices = mx.concatenate(
                    [
                        mx.full((r,), idx)
                        for idx, r in zip(input_indices.tolist(), repeats_arr)
                    ]
                )
                grad_input_flat = grad_input_flat.at[output_indices].add(
                    grad_output_flat
                )

            grad_input = grad_input_flat.reshape(a.shape)

        else:
            axis_ = self.axis % a.ndim
            if repeats_arr.size == 1:
                repeats_arr = lib_.full(a.shape[self.axis], repeats_arr)
            elif repeats_arr.size != a.shape[self.axis]:
                raise ValueError(
                    "repeats must be an integer or a "
                    + "sequence of the same length as the axis dimension."
                )

            expand_dims = [1] * a.ndim
            expand_dims[axis_] = -1

            input_indices_axis = lib_.arange(a.shape[axis_]).reshape(expand_dims)
            if lib_ is np:
                output_indices_axis = np.repeat(
                    input_indices_axis, repeats_arr, axis=axis_
                )
            else:
                moved_input_ = mx.moveaxis(input_indices_axis, axis_, 0)
                input_shape_ = input_indices_axis.shape
                slices = [
                    mx.full(
                        (*input_shape_[:axis_], r, *input_shape_[axis_ + 1 :]),
                        mx.expand_dims(slice_, axis_),
                    )
                    for slice_, r in zip(moved_input_, repeats_arr)
                ]
                output_indices_axis = mx.concatenate(slices, axis=axis_)

            idx = lib_.stack(
                lib_.meshgrid(
                    *[lib_.arange(s) for s in self.result.grad.shape], indexing="ij"
                )
            )
            idx[axis_] = output_indices_axis

            if lib_ is np:
                np.add.at(grad_input, tuple(idx), self.result.grad)
            else:
                grad_input = grad_input.at[tuple(idx)].add(self.result.grad)

        return grad_input


class tile(operation):
    def __init__(self, reps: int | Sequence[int]) -> None:
        super().__init__()
        self.reps = reps

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.tile(a.data, self.reps))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.tile(a.data, self.reps))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        if a.ndim == 0:
            input_shape = (1,)
            if isinstance(self.reps, int):
                reps_list = (self.reps,)
            else:
                reps_list = tuple(self.reps)
                if len(reps_list) == 0:
                    reps_list = (1,)
        else:
            input_shape = lib_.array(a.shape)
            if isinstance(self.reps, int):
                reps_list = (1,) * (a.ndim - 1) + (self.reps,)
            else:
                reps_list = tuple(self.reps)
                if len(reps_list) < self.ndim:
                    reps_list = (1,) * (a.ndim - len(reps_list)) + reps_list

        reps_array = lib_.array(reps_list)

        reshape_dims = []
        for dim_size, rep in zip(input_shape, reps_array):
            reshape_dims.extend([rep, dim_size])

        grad_output = self.result.grad
        if grad_output.size != lib_.prod(lib_.array(reshape_dims)):
            raise ValueError(
                f"Cannot reshape array of size {grad_output.size} "
                + f"into shape {reshape_dims}"
            )

        grad_output_reshape = grad_output.reshape(reshape_dims)
        axes_to_sum = tuple(range(0, grad_output_reshape.ndim, 2))

        return grad_output_reshape.sum(axis=axes_to_sum)


class flatten(operation):
    def __init__(self, start_axis: int = 0, end_axis: int = -1) -> None:
        super().__init__()
        self.start_axis = start_axis
        self.end_axis = end_axis

    def _unified(self, a: Tensor) -> _FuncOpReturnType:
        self.original_shape = a.shape

        start = self.start_axis if self.start_axis >= 0 else a.ndim + self.start_axis
        end = self.end_axis if self.end_axis >= 0 else a.ndim + self.end_axis

        flat_axis = 1
        for i in range(start, end + 1):
            flat_axis *= a.shape[i]

        new_shape = a.shape[:start] + (flat_axis,) + a.shape[end + 1 :]
        self.result = Tensor(a.data.reshape(new_shape))
        return self.result, self.__grad__

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a)

    def __grad__(self) -> _GradFuncType:
        return self.result.grad.reshape(self.original_shape)


class meshgrid(operation):
    def __init__(self, indexing: Literal["xy", "ij"]) -> None:
        super().__init__()
        if indexing not in {"xy", "ij"}:
            raise ValueError("indexing must be either 'xy' or 'ij'")
        self.indexing = indexing

    def _unified(self, a: Tensor, b: Tensor, lib_: ModuleType) -> tuple[Tensor, Tensor]:
        if a.ndim != 1 or a.ndim != 1:
            raise ValueError(f"Input tensors must be 1D tensors.")

        X = lib_.repeat(a.data.reshape(1, -1), b.data.shape[0], axis=0)
        Y = lib_.repeat(b.data.reshape(-1, 1), a.data.shape[0], axis=1)

        if self.indexing == "xy":
            X, Y = Y, X

        return Tensor(X), Tensor(Y)

    @func_op(n_in=2, n_ret=2)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.X, self.Y = self._unified(a, b, lib_=np)
        self.result = (self.X, self.Y)

        compute_grad = partial(self.__grad__, lib_=np)
        return (self.X, compute_grad), (self.Y, compute_grad)

    @func_op(n_in=2, n_ret=2, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.X, self.Y = self._unified(a, b, lib_=mx)
        self.result = (self.X, self.Y)

        compute_grad = partial(self.__grad__, lib_=mx)
        return (self.X, compute_grad), (self.Y, compute_grad)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        grad_x = lib_.sum(self.X.grad, axis=0)
        grad_y = lib_.sum(self.Y.grad, axis=1)

        return grad_x, grad_y


class split(operation):
    def __init__(
        self, size_or_sections: int | list[int] | tuple[int], axis: int
    ) -> None:
        super().__init__()
        self.size_or_sections = size_or_sections
        self.axis = axis

    def cpu(self, *args, **kwargs) -> _FuncOpReturnType:
        return super().cpu(*args, **kwargs)

    def gpu(self, *args, **kwargs) -> _FuncOpReturnType:
        return super().gpu(*args, **kwargs)

    @override
    def __call__(self, a: Tensor) -> tuple[Tensor, ...]:
        returns = []
        if self.axis < 0:
            self.axis = a.ndim + self.axis

        self.axislen = a.shape[self.axis]
        if isinstance(self.size_or_sections, int):
            self.size_or_sections = (self.size_or_sections,) * int(
                math.ceil(self.axislen / self.size_or_sections)
            )

        cur_idx = 0
        for size in self.size_or_sections:
            slices = []
            for _ in range(self.axis):
                slices.append(slice(None, None, None))

            slices.append(slice(cur_idx, cur_idx + size, None))
            returns.append(a[*slices])
            cur_idx += size

        self.result = tuple(returns)
        return self.result


class tril(operation):
    def __init__(self, diagonal: int) -> None:
        super().__init__()
        self.diagonal = diagonal

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.tril(a.data, k=self.diagonal))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.tril(a.data, k=self.diagonal))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.tril(self.result.grad, k=self.diagonal)


class triu(operation):
    def __init__(self, diagonal: int) -> None:
        super().__init__()
        self.diagonal = diagonal

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.triu(a.data, k=self.diagonal))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.triu(a.data, k=self.diagonal))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.triu(self.result.grad, k=self.diagonal)


class broadcast_to(operation):
    def __init__(self, shape: _ShapeLike) -> None:
        super().__init__()
        self.shape = shape

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.original_shape = a.shape
        self.result = Tensor(np.broadcast_to(a.data, self.shape))

        return self.result, self.__grad__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.original_shape = a.shape
        self.result = Tensor(mx.broadcast_to(a.data, self.shape))

        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        input_shape = self.original_shape
        ndim_diff = len(self.shape) - len(input_shape)
        if ndim_diff > 0:
            input_shape = (1,) * ndim_diff + input_shape

        for axis, (in_dim, out_dim) in enumerate(zip(input_shape, self.shape)):
            if in_dim == 1 and out_dim > 1:
                self.result.grad = self.result.grad.sum(axis=axis, keepdims=True)

        return self.result.grad.reshape(self.original_shape)


class chunk(operation):
    def __init__(self, chunks: int, axis: int) -> None:
        super().__init__()
        self.chunks = chunks
        self.axis = axis

    def _unified(self, a: Tensor, lib_: ModuleType) -> _FuncOpReturnType:
        if self.chunks <= 0:
            raise ValueError("chunks must be greater than 0.")

        dim_size = a.shape[self.axis]
        chunk_size = (dim_size + self.chunks - 1) // self.chunks

        split_indices = list(range(chunk_size, dim_size, chunk_size))
        chunked_arrays = lib_.split(a.data, split_indices, axis=self.axis)

        results = []
        start_idx = 0
        for arr in chunked_arrays:
            chunk_t = Tensor(arr)

            def compute_grad(_a: Tensor = chunk_t, _idx=start_idx) -> _GradFuncType:
                slices = [slice(None)] * a.ndim
                slices[self.axis] = slice(_idx, _idx + _a.shape[self.axis])

                grad = lib_.zeros_like(a.data)
                grad[tuple(slices)] = _a.grad

                return grad

            results.append((chunk_t, compute_grad))
            start_idx += chunk_t.shape[self.axis]

        return tuple(results)

    @func_op(n_in=1, n_ret=None)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a, lib_=np)

    @func_op(n_in=1, n_ret=None, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a, lib_=mx)


class masked_fill(operation):
    def __init__(self, mask: Tensor, value: _Scalar) -> None:
        super().__init__()
        self.mask = mask
        self.value = value

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.where(self.mask.data.astype(bool), self.value, a.data))
        return self.result, self.__grad_cpu__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.where(self.mask.data.astype(bool), self.value, a.data))
        return self.result, self.__grad_gpu__

    def __grad_cpu__(self) -> _GradFuncType:
        grad = self.result.grad.copy()
        grad[self.mask.data] = 0
        return grad

    def __grad_gpu__(self) -> _GradFuncType:
        grad = mx.array(self.result.grad)
        grad = mx.where(self.mask.data.astype(bool), 0, grad)
        return grad


class roll(operation):
    def __init__(
        self, shifts: int | tuple[int, ...], axis: int | tuple[int, ...] | None
    ) -> None:
        super().__init__()
        self.shifts = shifts
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.roll(a.data, shift=self.shifts, axis=self.axis))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.roll(a.data, shift=self.shifts, axis=self.axis))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        if isinstance(self.shifts, int):
            neg_shift = -self.shifts
        elif isinstance(self.shifts, tuple):
            neg_shift = tuple(-s for s in self.shifts)

        return lib_.roll(self.result.grad, shift=neg_shift, axis=self.axis)


class unbind(operation):
    def __init__(self, axis: int = 0) -> None:
        super().__init__()
        self.axis = axis

    def _take_along_axis(self, x: _ArrayLike, index: int) -> _ArrayLike:
        idx = [slice(None)] * x.ndim
        idx[self.axis] = index
        return x[tuple(idx)]

    def _unified(self, a: Tensor, lib_: ModuleType) -> _FuncOpReturnType:
        dim_size = a.shape[self.axis]
        results = []

        for idx in range(dim_size):
            arr = self._take_along_axis(a.data, idx)
            ten = Tensor(arr)

            def compute_grad(_a: Tensor = ten, _idx: int = idx) -> _GradFuncType:
                idx_slices = [slice(None)] * a.ndim
                idx_slices[self.axis] = _idx

                grad = lib_.zeros_like(a.data)
                grad[tuple(idx_slices)] = _a.grad
                return grad

            results.append((ten, compute_grad))

        return tuple(results)

    @func_op(n_in=1, n_ret=None)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a, lib_=np)

    @func_op(n_in=1, n_ret=None, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._unified(a, lib_=mx)


_SortKind = Literal["quicksort", "mergesort", "heapsort", "stable"]


class sort(operation):
    def __init__(
        self,
        axis: int = -1,
        descending: bool = False,
        kind: _SortKind | None = None,
        stable: bool = False,
    ) -> None:
        super().__init__()
        self.axis = axis
        self.descending = descending
        self.stable = stable

        kind = kind or "quicksort"
        if self.stable:
            kind = "stable"
        self.kind = kind

    @func_op(n_in=1, n_ret=2)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        data = a.data
        indices = np.argsort(data, axis=self.axis, kind=self.kind)
        sorted_data = np.take_along_axis(data, indices, axis=self.axis)

        if self.descending:
            indices = np.flip(indices, axis=self.axis)
            sorted_data = np.take_along_axis(data, indices, axis=self.axis)

        values = Tensor(sorted_data)
        indices_t = Tensor(indices)

        self.result = (values, indices_t)
        return (
            (values, partial(self.__grad__, lib_=np)),
            (indices_t, lambda: np.zeros_like(indices)),
        )

    @func_op(n_in=1, n_ret=2, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        data = a.data
        indices = mx.argsort(data, axis=self.axis)

        if self.descending:
            shape = [1] * data.ndim
            shape[self.axis] = -1

            rev_idx = mx.arange(indices.shape[self.axis] - 1, -1, -1)
            rev_idx = rev_idx.reshape(*shape)
            indices = mx.take_along_axis(indices, rev_idx, axis=self.axis)

        sorted_data = mx.take_along_axis(data, indices, axis=self.axis)

        values = Tensor(sorted_data)
        indices_t = Tensor(indices)

        self.result = (values, indices_t)
        return (
            (values, partial(self.__grad__, lib_=mx)),
            (indices_t, lambda: mx.zeros_like(indices)),
        )

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        grad = self.result[0].grad
        reverse_indices = lib_.argsort(self.result[1].data, axis=self.axis)

        grad_out = lib_.take_along_axis(grad, reverse_indices, axis=self.axis)
        return grad_out

    def __flops__(self, a: Tensor) -> int:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        n = a.shape[axis]

        num_slices = math.prod([s for i, s in enumerate(a.shape) if i != axis])
        return int(num_slices * n * math.log2(max(n, 2)))


class argsort(operation):
    def __init__(
        self,
        axis: int = -1,
        descending: bool = False,
        kind: _SortKind | None = None,
        stable: bool = False,
    ) -> None:
        super().__init__()
        self.axis = axis
        self.descending = descending
        self.stable = stable

        kind = kind or "quicksort"
        if self.stable:
            kind = "stable"
        self.kind = kind

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        data = -a.data if self.descending else a.data
        indices = np.argsort(data, axis=axis, kind=self.kind)

        self.result = Tensor(indices.astype(np.int32))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        data = mx.negative(a.data) if self.descending else a.data
        indices = mx.argsort(data, axis=axis)

        self.result = Tensor(indices.astype(mx.int32))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        n = a.shape[axis]
        num_slices = math.prod([s for i, s in enumerate(a.shape) if i != axis])
        return int(num_slices * n * math.log2(max(n, 2)))


class argmin(operation):
    def __init__(self, axis: int = None, keepdims: bool = False) -> None:
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis is not None else 0
        indices = np.argmin(a.data, axis=axis)
        if self.keepdims:
            indices = np.expand_dims(indices, axis)
        self.result = Tensor(indices.astype(np.int32))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis is not None else 0
        indices = mx.argmin(a.data, axis=axis)
        if self.keepdims:
            indices = mx.expand_dims(indices, axis)
        self.result = Tensor(indices.astype(mx.int32))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        axis = self.axis if self.axis is not None else 0
        num_slices = math.prod([s for i, s in enumerate(a.shape) if i != axis])
        return num_slices * a.shape[axis]


class argmax(operation):
    def __init__(self, axis: int | None = None, keepdims: bool = False) -> None:
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis is not None else 0
        indices = np.argmax(a.data, axis=axis)
        if self.keepdims:
            indices = np.expand_dims(indices, axis)

        self.result = Tensor(indices.astype(np.int32))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.aixs if self.axis is not None else 0
        indices = mx.argmax(a.data, axis=axis)
        if self.keepdims:
            indices = mx.expand_dims(indices, axis)

        self.result = Tensor(indices.astype(mx.int32))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        axis = self.axis if self.axis is not None else 0
        num_slices = math.prod([s for i, s in enumerate(a.shape) if i != axis])
        return num_slices * a.shape[axis]


class nonzero(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        coords = np.transpose(np.nonzero(a.data))
        self.result = Tensor(coords.astype(np.int32))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        shape = a.shape
        ndim = a.ndim
        size = a.size

        mask = mx.not_equal(a.data, mx.zeros_like(a.data)).reshape(-1)
        flat_idx = mx.arange(size)

        marked = mx.where(mask, flat_idx, size)
        sorted_idx = mx.sort(marked)
        count = int(mx.sum(mx.less(sorted_idx, size)).item())
        kept = sorted_idx[:count]

        strides = []
        acc = 1
        for s in reversed(shape):
            strides.append(acc)
            acc *= s

        strides = strides[::-1]
        strides_tensor = mx.array(strides).reshape(1, ndim)

        kept = kept.reshape(-1, 1)
        coords = (kept // strides_tensor) % mx.array(shape).reshape(1, ndim)
        coords = coords.astype(mx.int32)

        self.result = Tensor(coords)
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        flat = a.data.flatten()
        return sum(bool(x) for x in flat)


@fallback
class unique(operation):
    def __init__(self, sorted: bool = True, axis: int | None = None) -> None:
        super().__init__()
        self.sorted = sorted
        self.axis = axis
        self.inverse_ = None

    def _unified(self, a: Tensor) -> _NumPyArray:
        data = a.data

        if self.sorted:
            unique_data, inverse = np.unique(data, axis=self.axis, return_inverse=True)
        else:
            unique_data, idx = np.unique(data, return_index=True, axis=self.axis)
            sorter = np.sort(idx)

            if self.axis is None:
                flat_data = np.ravel(data)
                unique_data = flat_data[sorter]
                _, inverse = np.unique(flat_data, return_inverse=True)
            else:
                unique_data = np.take(data, sorter, axis=self.axis)
                inverse = None

        self.inverse_ = None if inverse is None else Tensor(inverse)
        return unique_data

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        out = self._unified(a)
        self.result = Tensor(out)
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        out = self._unified(a)
        self.result = Tensor(mx.array(out))
        if self.inverse_ is not None:
            self.inverse_ = Tensor(mx.array(self.inverse_.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        return int(a.size * math.log2(max(a.size, 2)))


class topk(operation):
    def __init__(
        self, k: int, axis: int = -1, largest: bool = True, sorted: bool = True
    ) -> None:
        super().__init__()
        self.k = k
        self.axis = axis
        self.largest = largest
        self.sorted = sorted

    @func_op(n_in=1, n_ret=2)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        data = a.data.copy()
        negate_for_sort = not self.largest

        if negate_for_sort:
            data = -data

        indices = np.argpartition(data, -self.k, axis=axis)[..., -self.k :]
        values = np.take_along_axis(data, indices, axis=axis)

        if negate_for_sort:
            values = -values
            data = -data

        if self.sorted:
            sort_order = (
                np.argsort(-values, axis=axis)
                if self.largest
                else np.argsort(values, axis=axis)
            )
            values = np.take_along_axis(values, sort_order, axis=axis)
            indices = np.take_along_axis(indices, sort_order, axis=axis)

        values_t = Tensor(values)
        indices_t = Tensor(indices.astype(np.int32))

        self.result = (values_t, indices_t)
        return (
            (values_t, partial(self.__grad_cpu__, a=a)),
            (indices_t, lambda: np.zeros_like(indices)),
        )

    @func_op(n_in=1, n_ret=2, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        axis = self.axis if self.axis >= 0 else a.ndim + self.axis
        data = a.data
        negate_for_sort = not self.largest

        if negate_for_sort:
            data = mx.negative(data)

        indices = mx.argsort(data, axis=axis)
        indices = mx.take(indices, mx.arange(-self.k, 0), axis=axis)
        values = mx.take_along_axis(data, indices, axis=axis)

        if negate_for_sort:
            values = mx.negative(values)
            data = mx.negative(data)

        if self.sorted:
            sort_order = (
                mx.argsort(mx.negative(values), axis=axis)
                if self.largest
                else mx.argsort(values, axis=axis)
            )
            values = mx.take_along_axis(values, sort_order, axis=axis)
            indices = mx.take_along_axis(indices, sort_order, axis=axis)

        values_t = Tensor(values)
        indices_t = Tensor(indices.astype(mx.int32))

        self.result = (values_t, indices_t)
        return (
            (values_t, partial(self.__grad_gpu__, a=a)),
            (indices_t, lambda: mx.zeros_like(indices)),
        )

    def __grad_cpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result[0].grad
        axis = self.axis if self.axis >= 0 else grad.ndim + self.axis

        output_grad = np.zeros_like(a.data)
        np.put_along_axis(output_grad, self.result[1].data, grad, axis=axis)
        return output_grad

    def __grad_gpu__(self, a: Tensor) -> _GradFuncType:
        grad = self.result[0].grad
        axis = self.axis if self.axis >= 0 else grad.ndim + self.axis

        input_shape = a.shape
        indices = self.result[1].data

        B, k = indices.shape
        D = input_shape[axis]

        grad_exp = grad.reshape(B, k, 1)
        indices_exp = indices.reshape(B, k, 1)

        class_range = mx.arange(D).reshape(1, 1, D)
        one_hot = (indices_exp == class_range).astype(mx.float32)

        scatter = grad_exp * one_hot
        output = mx.sum(scatter, axis=1)
        if axis != 1:
            output = mx.moveaxis(output, 1, axis)

        return output

    def __flops__(self, a: Tensor) -> int:
        shape = a.shape
        axis = self.axis if self.axis >= 0 else len(shape) + self.axis

        n = shape[axis]
        num_slices = math.prod([s for i, s in enumerate(shape) if i != axis])

        if self.sorted:
            return int(num_slices * (n + self.k * math.log2(max(self.k, 1))))
        else:
            return int(num_slices * n)


class histogramdd(operation):
    def __init__(
        self,
        bins: int | list[int],
        range: list[tuple[float, float]],
        density: bool = False,
    ) -> None:
        super().__init__()
        if isinstance(bins, int):
            bins = [bins] * len(range)
        if not all(isinstance(b, int) for b in bins):
            raise TypeError("All elements of bins must be integers.")

        self.bins = bins
        self.range = range
        self.density = density

    @func_op(n_in=1, n_ret=2, has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        hist, edges = np.histogramdd(a.data, self.bins, self.range, self.density)

        hist_t = Tensor(hist)
        edges_t = Tensor(np.stack([e for e in edges]))
        self.result = (hist_t, edges_t)
        return (
            (hist_t, partial(self.__grad__, lib_=np)),
            (edges_t, partial(self.__grad__, lib_=np)),
        )

    @func_op(n_in=1, n_ret=2, has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        data = a.data
        N, D = data.shape[:2]
        bins = self.bins

        edges = [mx.linspace(*r, bins[i] + 1) for i, r in enumerate(self.range)]
        bin_widths = [(r[1] - r[0]) / bins[i] for i, r in enumerate(self.range)]

        bin_indices = []
        for d in range(D):
            col = data[:, d]
            idx = mx.floor((col - self.range[d][0]) / bin_widths[d]).astype(mx.int32)
            idx = mx.clip(idx, 0, bins[d] - 1)
            bin_indices.append(idx)

        bin_indices = mx.stack(bin_indices, axis=1)
        hist_shape = tuple(bins)
        hist = mx.zeros(hist_shape, dtype=mx.int32)

        for i in range(N):
            idx = tuple(int(bin_indices[i, d].item()) for d in range(D))
            hist[idx] += 1

        hist = hist.astype(mx.float32)
        if self.density:
            total = mx.sum(hist)
            bin_volume = math.prod(bin_widths)
            hist /= total * bin_volume

        hist_t = Tensor(hist)
        edges_t = Tensor(mx.stack([e for e in edges]))
        self.result = (hist_t, edges_t)
        return (
            (hist_t, partial(self.__grad__, lib_=np)),
            (edges_t, partial(self.__grad__, lib_=np)),
        )

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        return int(math.prod(a.shape) + math.prod(self.bins))


class where(operation):
    def __init__(self) -> None:
        super().__init__()

    @func_op(n_in=3, n_ret=1)
    def cpu(self, condition: Tensor, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.cond_ = condition
        self.result = Tensor(np.where(condition.data, a.data, b.data))
        return self.result, partial(self.__grad__, lib_=np)

    @func_op(n_in=3, n_ret=1, device="gpu")
    def gpu(self, condition: Tensor, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.cond_ = condition
        self.result = Tensor(mx.where(condition.data, a.data, b.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        cond = self.cond_.data
        grad = self.result.grad

        grad_cond = lib_.array(0.0)
        grad_a = lib_.where(cond, grad, 0)
        grad_b = lib_.where(lib_.logical_not(cond), grad, 0)

        return grad_cond, grad_a, grad_b

    def __flops__(self, condition: Tensor, a: Tensor, b: Tensor) -> int:
        return max(condition.size, a.size, b.size)


class diagonal(operation):
    def __init__(self, offset: int = 0, axis1: int = 0, axis2: int = 1) -> None:
        super().__init__()
        self.offset = offset
        self.axis1 = axis1
        self.axis2 = axis2

    @unary_func_op()
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.diagonal(a.data, self.offset, self.axis1, self.axis2))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.diagonal(a.data, self.offset, self.axis1, self.axis2))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_) -> _GradFuncType:
        grad_out = self.result.grad
        grad_in = lib_.zeros_like(a.data)

        if self.offset >= 0:
            size = min(a.shape[self.axis1], a.shape[self.axis2] - self.offset)
            i = lib_.arange(size, dtype=lib_.int32)
            j = i + self.offset
        else:
            size = min(a.shape[self.axis1] + self.offset, a.shape[self.axis2])
            i = lib_.arange(size, dtype=lib_.int32) - self.offset
            j = lib_.arange(size, dtype=lib_.int32)

        indexer = []
        for ax in range(a.ndim):
            if ax == self.axis1:
                indexer.append(i)
            elif ax == self.axis2:
                indexer.append(j)
            else:
                indexer.append(lib_.zeros_like(i))

        if lib_ is np:
            grad_in[tuple(indexer)] += grad_out
        else:
            grad_in = grad_in.at[tuple(indexer)].add(grad_out)

        return grad_in

    def __flops__(self, a: Tensor) -> int:
        return min(a.shape[self.axis1], a.shape[self.axis2])



import re
import numpy as np

from functools import partial
from types import ModuleType
from typing import Literal

from lucid._tensor import Tensor
from lucid.types import _EinopsPattern, _NumPyArray, _MLXArray, _ShapeLike

from lucid._backend.core import (
    operation,
    unary_func_op,
    poly_func_op,
    _FuncOpReturnType,
    _GradFuncType,
)
from lucid._backend.metal import mx


_ReduceStr = Literal["sum", "mean"]


def _parse_pattern(pattern_side: str) -> list[str | tuple[str, ...]]:
    parts = re.findall(r"\.\.\.|\([^)]+\)|\S+", pattern_side)
    tokens: list[str | tuple[str, ...]] = []

    for part in parts:
        if part.startswith("(") and part.endswith(")"):
            inner = part[1:-1].strip()
            if not inner:
                raise ValueError("Empty group in pattern.")
            tokens.append(tuple(inner.split()))
        else:
            tokens.append(part)

    return tokens


def _build_intermediate(
    input_tokens: list[str | tuple[str, ...]], shape: tuple[int, ...], shapes: dict
) -> tuple[list[str], list[int]]:
    inter_tokens: list[str] = []
    inter_shape: list[int] = []

    for token, dim in zip(input_tokens, shape):
        if isinstance(token, tuple):
            prod = 1
            group_sizes = []
            for t in token:
                if t not in shapes:
                    raise ValueError(
                        f"Size for token '{t}' in group {token} must be provided."
                    )
                s = shapes[t]
                group_sizes.append(s)
                prod *= s

            if prod != dim:
                raise ValueError(
                    f"Product of sizes {prod} for grouped tokens "
                    + f"{token} does not match merged axis size {dim}."
                )
            inter_tokens.extend(token)
            inter_shape.extend(group_sizes)

        else:
            if token in shapes and shapes[token] != dim:
                raise ValueError(
                    f"Provided size for token '{token}' ({shapes[token]}) "
                    + f"does not match tensor dimension ({dim})."
                )
            inter_tokens.append(token)
            inter_shape.append(dim)

    return inter_tokens, inter_shape


class rearrange(operation):
    def __init__(
        self, pattern: _EinopsPattern, t_shape: _ShapeLike, **shapes: int
    ) -> None:
        super().__init__()
        try:
            in_pat, out_pat = map(str.strip, pattern.split("->"))
        except Exception as e:
            raise ValueError(
                "Pattern must contain '->' separating input and output patterns."
            ) from e

        input_tokens = _parse_pattern(in_pat)
        output_tokens = _parse_pattern(out_pat)

        if "..." in input_tokens:
            if input_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in input pattern.")
            ell_pos = input_tokens.index("...")
            num_unnamed = len(t_shape) - (len(input_tokens) - 1)
            if num_unnamed < 1:
                raise ValueError("Ellipsis '...' expands to zero dimensions.")
            ell_names = [f"_ellipsis_{i}" for i in range(num_unnamed)]
            input_tokens = (
                input_tokens[:ell_pos] + ell_names + input_tokens[ell_pos + 1 :]
            )
        else:
            ell_names = []

        if "..." in output_tokens:
            if not ell_names:
                raise ValueError("Output pattern has '...' but input pattern did not.")
            if output_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in output pattern.")
            ell_pos_out = output_tokens.index("...")
            output_tokens = (
                output_tokens[:ell_pos_out]
                + ell_names
                + output_tokens[ell_pos_out + 1 :]
            )

        if len(input_tokens) != len(t_shape):
            raise ValueError(
                f"Input pattern has {len(input_tokens)} tokens, "
                + f"but tensor has {len(t_shape)} dimensions."
            )

        self.inter_tokens, self.inter_shape = _build_intermediate(
            input_tokens, t_shape, shapes
        )

        self.perm: list[int] = []
        self.group_splits: list[tuple[int, ...]] = []
        used = set()

        for token in output_tokens:
            if isinstance(token, tuple):
                group_perm, group_dims = [], []
                for t in token:
                    found = None
                    for i, it in enumerate(self.inter_tokens):
                        if it == t and i not in used:
                            found = i
                            break
                    if found is None:
                        raise ValueError(
                            f"Token '{t}' in output group {token} not found in input."
                        )
                    group_perm.append(found)
                    group_dims.append(self.inter_shape[found])
                    used.add(found)

                self.group_splits.append(tuple(group_dims))
                self.perm.extend(group_perm)

            else:
                found = None
                for i, it in enumerate(self.inter_tokens):
                    if it == token and i not in used:
                        found = i
                        break
                if found is None:
                    raise ValueError(
                        f"Token '{token}' from output pattern not found in input."
                    )

                self.perm.append(found)
                self.group_splits.append((self.inter_shape[found],))
                used.add(found)

        self.final_shape = tuple(np.prod(group).item() for group in self.group_splits)

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        inter = a.data.reshape(tuple(self.inter_shape))
        transposed = np.transpose(inter, axes=self.perm)

        self.result = Tensor(transposed.reshape(self.final_shape))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        inter = a.data.reshape(tuple(self.inter_shape))
        transposed = mx.transpose(inter, axes=self.perm)

        self.result = Tensor(transposed.reshape(self.final_shape))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        unmerged_shape = tuple(dim for group in self.group_splits for dim in group)
        grad_unmerged = self.result.grad.reshape(unmerged_shape)

        inv_perm = lib_.argsort(lib_.array(self.perm))
        grad_interm = lib_.transpose(grad_unmerged, axes=inv_perm)

        return grad_interm.reshape(a.shape)


class reduce(operation):
    def __init__(
        self,
        pattern: _EinopsPattern,
        reduction: _ReduceStr,
        t_shape: _ShapeLike,
        **shapes: int,
    ) -> None:
        super().__init__()
        self.reduction = reduction
        try:
            in_pat, out_pat = map(str.strip, pattern.split("->"))
        except Exception as e:
            raise ValueError(
                "Pattern must contain '->' separating input and output patterns."
            ) from e

        input_tokens = _parse_pattern(in_pat)
        output_tokens = _parse_pattern(out_pat)

        if "..." in input_tokens:
            if input_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in input pattern.")
            ell_pos = input_tokens.index("...")
            num_unnamed = len(t_shape) - (len(input_tokens) - 1)
            if num_unnamed < 1:
                raise ValueError("Ellipsis '...' expands to zero dimensions.")
            ell_names = [f"_ellipsis_{i}" for i in range(num_unnamed)]
            input_tokens = (
                input_tokens[:ell_pos] + ell_names + input_tokens[ell_pos + 1 :]
            )
        else:
            ell_names = []

        if "..." in output_tokens:
            if not ell_names:
                raise ValueError("Output pattern has '...' but input pattern did not.")
            if output_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in output pattern.")
            ell_pos_out = output_tokens.index("...")
            output_tokens = (
                output_tokens[:ell_pos_out]
                + ell_names
                + output_tokens[ell_pos_out + 1 :]
            )

        if len(input_tokens) != len(t_shape):
            raise ValueError(
                f"Input pattern has {len(input_tokens)} tokens, "
                + f"but tensor has {len(t_shape)} dimensions."
            )

        self.inter_tokens, self.inter_shape = _build_intermediate(
            input_tokens, t_shape, shapes
        )
        kept_indices: list[int] = []
        kept_groups: list[tuple[int, ...]] = []
        used = set()

        for token in output_tokens:
            if isinstance(token, tuple):
                group_inds, group_dims = [], []
                for t in token:
                    found = None
                    for i, it in enumerate(self.inter_tokens):
                        if it == t and i not in used:
                            found = i
                            break
                    if found is None:
                        raise ValueError(
                            f"Token '{t}' in output group {token} not found in input."
                        )

                    group_inds.append(found)
                    group_dims.append(self.inter_shape[found])
                    used.add(found)

                kept_indices.extend(group_inds)
                kept_groups.append(tuple(group_dims))

            else:
                found = None
                for i, it in enumerate(self.inter_tokens):
                    if it == token and i not in used:
                        found = i
                        break
                if found is None:
                    raise ValueError(
                        f"Token '{token}' from output pattern not found in input."
                    )

                kept_indices.append(found)
                kept_groups.append((self.inter_shape[found],))
                used.add(found)

        all_indices = set(range(len(self.inter_tokens)))
        reduced_indices = sorted(list(all_indices - used))

        self.perm = kept_indices + reduced_indices

        self.kept_flat_shape = tuple(self.inter_shape[i] for i in kept_indices)
        self.reduced_shape = tuple(self.inter_shape[i] for i in reduced_indices)
        self.final_shape = tuple(np.prod(group) for group in kept_groups)

    def _unified(self, arr: _NumPyArray | _MLXArray) -> _NumPyArray | _MLXArray:
        intermediate = arr.reshape(tuple(self.inter_shape))
        transposed = intermediate.transpose(*self.perm)

        reduce_axes = tuple(
            range(
                len(self.kept_flat_shape),
                len(self.kept_flat_shape) + len(self.reduced_shape),
            )
        )
        if self.reduction == "sum":
            reduced_data = transposed.sum(axis=reduce_axes)
        elif self.reduction == "mean":
            reduced_data = transposed.mean(axis=reduce_axes)
        else:
            raise ValueError(f"Unsupported reduction method: {self.reduction}")

        return reduced_data.reshape(self.final_shape)

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(self._unified(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(self._unified(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad_kept = self.result.grad.reshape(self.kept_flat_shape)
        new_shape = self.kept_flat_shape + (1,) * len(self.reduced_shape)

        grad_expanded = grad_kept.reshape(new_shape)
        grad_broadcast = lib_.broadcast_to(
            grad_expanded, self.kept_flat_shape + tuple(self.reduced_shape)
        )

        if self.reduction == "mean":
            factor = (
                lib_.prod(lib_.array(self.reduced_shape)) if self.reduced_shape else 1
            )
            grad_broadcast = grad_broadcast / factor

        inv_perm = lib_.argsort(lib_.array(self.perm))
        grad_intermediate = lib_.transpose(grad_broadcast, axes=inv_perm)

        return grad_intermediate.reshape(a.shape)

    def __flops__(self, _) -> int:
        reduced_size = np.prod(self.reduced_shape)
        output_size = np.prod(self.final_shape)

        if self.reduction == "sum":
            return output_size * (reduced_size - 1)
        elif self.reduction == "mean":
            return output_size * reduced_size
        else:
            raise ValueError(f"Unsupported reduction type: {self.reduction}")


class repeat(operation):
    def __init__(
        self, pattern: _EinopsPattern, t_shape: _ShapeLike, **shapes: int
    ) -> None:
        super().__init__()
        try:
            in_pat, out_pat = map(str.strip, pattern.split("->"))
        except Exception as e:
            raise ValueError(
                "Pattern must contain '->' separating input and output patterns."
            ) from e

        input_tokens = _parse_pattern(in_pat)
        output_tokens = _parse_pattern(out_pat)

        if "..." in input_tokens:
            if input_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in input pattern.")
            ell_pos = input_tokens.index("...")
            num_unnamed = len(t_shape) - (len(input_tokens) - 1)
            if num_unnamed < 1:
                raise ValueError("Ellipsis '...' expands to zero dimensions.")
            ell_names = [f"_ellipsis_{i}" for i in range(num_unnamed)]
            input_tokens = (
                input_tokens[:ell_pos] + ell_names + input_tokens[ell_pos + 1 :]
            )
        else:
            ell_names = []

        if "..." in output_tokens:
            if not ell_names:
                raise ValueError("Output pattern has '...' but input pattern did not.")
            if output_tokens.count("...") > 1:
                raise ValueError("Only one ellipsis '...' allowed in output pattern.")
            ell_pos_out = output_tokens.index("...")
            output_tokens = (
                output_tokens[:ell_pos_out]
                + ell_names
                + output_tokens[ell_pos_out + 1 :]
            )

        if len(input_tokens) != len(t_shape):
            raise ValueError(
                f"Input pattern has {len(input_tokens)} tokens, "
                + f"but tensor has {len(t_shape)} dimensions."
            )

        intermediate_tokens, self.intermediate_shape = _build_intermediate(
            input_tokens, t_shape, shapes
        )

        used_indices = set()
        group_splits: list[tuple[int, ...]] = []
        perm: list[int] = []

        for token in output_tokens:
            if isinstance(token, tuple):
                group_perm, group_dims = [], []
                for t in token:
                    found = None
                    for i, it in enumerate(intermediate_tokens):
                        if it == t and i not in used_indices:
                            found = i
                            break

                    if found is None:
                        group_perm.append(-1)
                        group_dims.append(1)
                    else:
                        group_perm.append(found)
                        group_dims.append(self.intermediate_shape[found])
                        used_indices.add(found)

                group_splits.append(tuple(group_dims))
                perm.extend(group_perm)

            else:
                found = None
                for i, it in enumerate(intermediate_tokens):
                    if it == token and i not in used_indices:
                        found = i
                        break

                if found is None:
                    perm.append(-1)
                    group_splits.append((1,))
                else:
                    perm.append(found)
                    group_splits.append((self.intermediate_shape[found],))
                    used_indices.add(found)

        self.base_shape = tuple(dim for group in group_splits for dim in group)
        out_shape_list = []
        idx = 0

        for token in output_tokens:
            if isinstance(token, tuple):
                n = len(token)
                group_perm = perm[idx : idx + n]

                if all(p == -1 for p in group_perm):
                    prod_val = 1
                    for t in token:
                        if t not in shapes:
                            raise ValueError(
                                f"Size for expansion token '{t}' must be provided."
                            )
                        prod_val *= shapes[t]
                    out_shape_list.append(prod_val)
                else:
                    prod_val = np.prod(self.base_shape[idx : idx + n])
                    out_shape_list.append(prod_val)

                idx += n
            else:
                if perm[idx] == -1:
                    out_shape_list.append(shapes[token])
                else:
                    out_shape_list.append(self.base_shape[idx])
                idx += 1

        self.out_shape = tuple(out_shape_list)
        self.kept_order = [p for p in perm if p != -1]

    def _unified(
        self, arr: _NumPyArray | _MLXArray, lib_: ModuleType
    ) -> _NumPyArray | _MLXArray:
        intermediate = arr.reshape(tuple(self.intermediate_shape))
        transposed = (
            lib_.transpose(intermediate, axes=self.kept_order)
            if self.kept_order
            else intermediate
        )
        base = transposed.reshape(self.base_shape)
        tile_multiples = tuple(o // b for b, o in zip(self.base_shape, self.out_shape))

        result_data = lib_.tile(base, tile_multiples)
        return result_data

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(self._unified(a.data, lib_=np))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(self._unified(a.data, lib_=mx))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        grad = self.result.grad
        for i, (b, o) in enumerate(zip(self.base_shape, self.out_shape)):
            if b == 1 and o != 1:
                grad = grad.sum(axis=i, keepdims=True)

        return grad.reshape(self.base_shape).reshape(a.shape)

    def __flops__(self, _) -> int:
        return int(np.prod(self.out_shape))


class einsum(operation):
    def __init__(self, pattern: str) -> None:
        super().__init__()
        self.pattern = pattern

    def _parse_equation(self, n_in: int) -> tuple[list[str], str]:
        eq = self.pattern.replace(" ", "")
        if "->" in eq:
            lhs, rhs = eq.split("->")
        else:
            lhs, rhs = eq, None

        inputs = lhs.split(",")
        if len(inputs) != n_in:
            raise ValueError(f"Pattern expects {len(inputs)} operands, got {n_in}.")

        if rhs is None:
            counts: dict[str, int] = {}
            for s in inputs:
                for c in s.replace("...", ""):
                    counts[c] = counts.get(c, 0) + 1

            rhs = "".join(sorted([c for c, v in counts.items() if v == 1]))
            if any("..." in s for s in inputs):
                rhs = "..." + rhs

        return inputs, rhs

    @poly_func_op()
    def cpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data for a in arr]
        self.in_sub, self.out_sub = self._parse_equation(len(arr))

        self.result = Tensor(np.einsum(self.pattern, *data_arr))
        return self.result, partial(self.__grad__, arr=arr, lib_=np)

    @poly_func_op(device="gpu")
    def gpu(self, *arr: Tensor) -> _FuncOpReturnType:
        data_arr = [a.data.astype(mx.float32) for a in arr]
        self.in_sub, self.out_sub = self._parse_equation(len(arr))

        self.result = Tensor(mx.einsum(self.pattern, *data_arr))
        return self.result, partial(self.__grad__, arr=arr, lib_=mx)

    def __grad__(self, arr: tuple[Tensor, ...], lib_: ModuleType) -> _GradFuncType:
        grads = []
        for i in range(len(arr)):
            other = [t.data for j, t in enumerate(arr) if j != i]

            eq = f"{self.out_sub}," + ",".join(self.in_sub[:i] + self.in_sub[i + 1 :])
            eq += f"->{self.in_sub[i]}"

            grads.append(lib_.einsum(eq, self.result.grad, *other))

        return tuple(grads)



import numpy as np

from lucid._tensor import Tensor
from lucid import types
from lucid.types import (
    _ShapeLike,
    _ArrayLike,
    _Scalar,
    _DeviceType,
    _NumPyArray,
    _MLXArray,
    _BuiltinNumeric,
    Numeric,
)


def _get_tensor_specs(
    a: Tensor | _ArrayLike,
    dtype: _BuiltinNumeric | Numeric | None,
    device: _DeviceType | None,
) -> tuple[_ArrayLike, type, _DeviceType]:
    if isinstance(a, Tensor):
        data = a.data
        dtype = a.dtype if dtype is None else dtype
        device = a.device if device is None else device
    else:
        data = a
        dtype = types.to_numeric_type(a.dtype)
        if isinstance(a, _NumPyArray):
            device = "cpu" if device is None else device
        elif isinstance(a, _MLXArray):
            device = "gpu" if device is None else device
        else:
            raise ValueError(f"Unknown input: '{type(a).__name__}'")

    return data, dtype, device


def zeros(
    shape: _ShapeLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.zeros(shape), requires_grad, keep_grad, dtype, device)


def zeros_like(
    a: Tensor | _ArrayLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType | None = None,
) -> Tensor:
    data, dtype, device = _get_tensor_specs(a, dtype, device)
    return Tensor(np.zeros_like(data), requires_grad, keep_grad, dtype, device)


def ones(
    shape: _ShapeLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.ones(shape), requires_grad, keep_grad, dtype, device)


def ones_like(
    a: Tensor | _ArrayLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType | None = None,
) -> Tensor:
    data, dtype, device = _get_tensor_specs(a, dtype, device)
    return Tensor(np.ones_like(data), requires_grad, keep_grad, dtype, device)


def eye(
    N: int,
    M: int | None = None,
    k: int = 0,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.eye(N, M, k), requires_grad, keep_grad, dtype, device)


def diag(
    v: Tensor | _ArrayLike,
    k: int = 0,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType | None = None,
) -> Tensor:
    data, dtype, device = _get_tensor_specs(v, dtype, device)
    return Tensor(np.diag(data, k=k), requires_grad, keep_grad, dtype, device)


def arange(
    start: _Scalar,
    stop: _Scalar,
    step: _Scalar,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.arange(start, stop, step), requires_grad, keep_grad, dtype, device)


def empty(
    shape: int | _ShapeLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.empty(shape), requires_grad, keep_grad, dtype, device)


def empty_like(
    a: Tensor | _ArrayLike,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType | None = None,
) -> Tensor:
    data, dtype, device = _get_tensor_specs(a, dtype, device)
    return Tensor(np.empty_like(data), requires_grad, keep_grad, dtype, device)


def linspace(
    start: _Scalar,
    stop: _Scalar,
    num: int = 50,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(
        np.linspace(start, stop, num), requires_grad, keep_grad, dtype, device
    )


def full(
    shape: int | _ShapeLike,
    fill_value: _Scalar,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.full(shape, fill_value), requires_grad, keep_grad, dtype, device)


def full_like(
    a: Tensor | _ArrayLike,
    fill_value: _Scalar,
    dtype: _BuiltinNumeric | Numeric | None = None,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType | None = None,
) -> Tensor:
    data, dtype, device = _get_tensor_specs(a, dtype, device)
    return Tensor(
        np.full_like(data, fill_value), requires_grad, keep_grad, dtype, device
    )



import math
import lucid
import lucid.nn.functional as F

from lucid._tensor import Tensor


def scaled_dot_product_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    attn_mask: Tensor | None = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    scale: float | None = None,
) -> Tensor:
    L, S = query.shape[-2], key.shape[-2]
    scale_factor = 1 / math.sqrt(query.shape[-1]) if scale is None else scale
    attn_bias = lucid.zeros(L, S, dtype=query.dtype).free()

    if is_causal:
        assert attn_mask is None
        temp_mask = 1 - lucid.ones(L, S).tril()
        attn_bias += temp_mask * -1e12

    if attn_mask is not None:
        attn_bias += attn_mask

    attn_weight = query @ key.mT * scale_factor
    attn_weight += attn_bias.broadcast_to(attn_weight.shape)
    attn_weight = F.softmax(attn_weight, axis=-1)
    attn_weight = F.dropout(attn_weight, dropout_p)

    return attn_weight @ value



import copy
from typing import Any, Iterable

import lucid.nn as nn
import lucid.optim as optim

from lucid._tensor import Tensor
from lucid.types import _OptimClosure


__all__ = ["SGD", "ASGD"]


class SGD(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1e-3,
        momentum: float = 0.0,
        weight_decay: float = 0.0,
    ) -> None:
        defaults = dict(lr=lr, momentum=momentum, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> Any | None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            momentum = group.get("momentum", self.defaults["momentum"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0:
                    grad = grad + weight_decay * param.data

                if momentum != 0:
                    param_state = self.state[param]
                    if "momentum_buffer" not in param_state:
                        buf = param_state["momentum_buffer"] = Tensor.copy_grad(
                            param.grad
                        )
                    else:
                        buf = param_state["momentum_buffer"]
                        buf = momentum * buf + grad
                        param_state["momentum_buffer"] = buf

                    grad = buf

                param.data = param.data - lr * grad

        return loss


class ASGD(optim.Optimizer):
    def __init__(
        self,
        params: Iterable[nn.Parameter],
        lr: float = 1e-3,
        momentum: float = 0.0,
        weight_decay: float = 0.0,
        alpha: float = 0.75,
        t0: float = 1e6,
        lambd: float = 1e-4,
    ) -> None:
        defaults = dict(
            lr=lr,
            momentum=momentum,
            weight_decay=weight_decay,
            alpha=alpha,
            t0=t0,
            lambd=lambd,
        )
        super().__init__(params, defaults)

    def step(self, closure: _OptimClosure | None = None) -> Any | None:
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group.get("lr", self.defaults["lr"])
            momentum = group.get("momentum", self.defaults["momentum"])
            weight_decay = group.get("weight_decay", self.defaults["weight_decay"])
            alpha = group.get("alpha", self.defaults["alpha"])
            t0 = group.get("t0", self.defaults["t0"])
            lambd = group.get("lambd", self.defaults["lambd"])

            for param in group["params"]:
                if param.grad is None:
                    continue

                grad = Tensor.copy_grad(param.grad)
                if weight_decay != 0:
                    grad = grad + weight_decay * param.data
                state = self.state[param]

                if len(state) == 0:
                    state["step"] = 0
                    if momentum != 0:
                        state["momentum_buffer"] = Tensor.copy_grad(grad)
                    state["ax"] = Tensor.copy_data(param.data)

                state["step"] += 1
                if momentum != 0:
                    buf = state["momentum_buffer"]
                    buf = momentum * buf + grad
                    state["momentum_buffer"] = buf
                    grad = buf

                param.data = param.data - lr * grad

                step = state["step"]
                if step >= t0:
                    ax = state["ax"]
                    coef = 1.0 / (alpha * step + 1)
                    ax = (1.0 - coef) * ax + coef * param.data - lambd * ax
                    state["ax"] = ax

        return loss

    def get_averages(self) -> Iterable[nn.Parameter]:
        averages = []
        for group in self.param_groups:
            for param in group["params"]:
                state = self.state[param]
                if "ax" in state:
                    avg_param = state["ax"]
                    averages.append(avg_param)
                else:
                    averages.append(param.data)

        return averages



import lucid
import lucid.nn as nn
from lucid._tensor import Tensor

from lucid.types import _ShapeLike


def normalize(
    input_: Tensor, ord: int = 2, axis: int = 1, eps: float = 1e-12
) -> Tensor:
    norm = lucid.sum(lucid.abs(input_) ** ord, axis=axis, keepdims=True) ** (1 / ord)
    norm = lucid.maximum(norm, lucid.tensor(eps))

    return input_ / norm


def batch_norm(
    input_: Tensor,
    running_mean: Tensor | None,
    running_var: Tensor | None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    spatial_dim = input_.ndim - 2

    use_batch_stats = training or running_mean is None or running_var is None

    if use_batch_stats:
        batch_mean = input_.mean(axis=(0, *range(2, input_.ndim)), keepdims=True)
        batch_var = input_.var(axis=(0, *range(2, input_.ndim)), keepdims=True)

        if running_mean is not None and running_var is not None:
            running_stats_ = (
                momentum * batch_mean.flatten() + (1 - momentum) * running_mean,
                momentum * batch_var.flatten() + (1 - momentum) * running_var,
            )

            if isinstance(running_mean, nn.Buffer) and isinstance(
                running_var, nn.Buffer
            ):
                running_mean.data = running_stats_[0].data
                running_var.data = running_stats_[1].data
            else:
                running_mean, running_var = running_stats_

        mean = batch_mean
        var = batch_var
    else:
        mean = running_mean.reshape(1, C, *(1,) * spatial_dim)
        var = running_var.reshape(1, C, *(1,) * spatial_dim)

    normalized = (input_ - mean) / lucid.sqrt(var + eps)

    if weight is not None:
        weight = weight.reshape((1, C) + (1,) * spatial_dim)
        normalized *= weight

    if bias is not None:
        bias = bias.reshape((1, C) + (1,) * spatial_dim)
        normalized += bias

    return normalized


def layer_norm(
    input_: Tensor,
    normalized_shape: _ShapeLike,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    eps: float = 1e-5,
) -> Tensor:
    if input_.shape[-len(normalized_shape) :] != normalized_shape:
        raise ValueError(
            "Input tensor's normalized shape must match "
            + "the provided `normalized_shape`."
        )

    mean = input_.mean(axis=tuple(range(-len(normalized_shape), 0)), keepdims=True)
    var = input_.var(axis=tuple(range(-len(normalized_shape), 0)), keepdims=True)

    normalized = (input_ - mean) / lucid.sqrt(var + eps)
    if weight is not None:
        normalized *= weight.reshape(
            (1,) * (input_.ndim - len(normalized_shape)) + normalized_shape
        )
    if bias is not None:
        normalized += bias.reshape(
            (1,) * (input_.ndim - len(normalized_shape)) + normalized_shape
        )

    return normalized


def instance_norm(
    input_: Tensor,
    running_mean: Tensor | None,
    running_var: Tensor | None,
    weight: Tensor | None = None,
    bias: Tensor | None = None,
    training: bool = True,
    momentum: float = 0.1,
    eps: float = 1e-5,
) -> Tensor:
    C = input_.shape[1]
    spatial_dims = input_.shape[2:]

    if training:
        axes = tuple(range(2, input_.ndim))
        instance_mean = input_.mean(axis=axes, keepdims=True)
        instance_var = input_.var(axis=axes, keepdims=True)

        if running_mean is not None and running_var is not None:
            running_stats_ = (
                momentum * instance_mean.mean(axis=0).flatten()
                + (1 - momentum) * running_mean,
                momentum * instance_var.mean(axis=0).flatten()
                + (1 - momentum) * running_var,
            )

            if isinstance(running_mean, nn.Buffer) and isinstance(
                running_var, nn.Buffer
            ):
                running_mean.data = running_stats_[0].data
                running_var.data = running_stats_[1].data
            else:
                running_mean, running_var = running_stats_

        mean = instance_mean
        var = instance_var
    else:
        mean = running_mean.reshape(1, C, *(1,) * len(spatial_dims))
        var = running_var.reshape(1, C, *(1,) * len(spatial_dims))

    normalized = (input_ - mean) / lucid.sqrt(var + eps)

    if weight is not None:
        weight = weight.reshape((1, C) + (1,) * len(spatial_dims))
        normalized *= weight

    if bias is not None:
        bias = bias.reshape((1, C) + (1,) * len(spatial_dims))
        normalized += bias

    return normalized


def group_norm(
    input_: Tensor,
    num_groups: int,
    weight: Tensor | None,
    bias: Tensor | None,
    eps: float = 1e-5,
) -> Tensor:
    N, C, *spatial_dims = input_.shape
    assert C % num_groups == 0, "Number of channels must be divisible by num_groups"

    group_size = C // num_groups
    new_shape = (N, num_groups, group_size, *spatial_dims)
    reshaped = input_.reshape(*new_shape)

    axes = (2,) + tuple(range(3, reshaped.ndim))
    mean = reshaped.mean(axis=axes, keepdims=True)
    var = reshaped.var(axis=axes, keepdims=True)

    normalized = (reshaped - mean) / lucid.sqrt(var + eps)
    normalized = normalized.reshape(N, C, *spatial_dims)

    if weight is not None:
        weight = weight.reshape(1, C, *(1,) * len(spatial_dims))
        normalized *= weight

    if bias is not None:
        bias = bias.reshape(1, C, *(1,) * len(spatial_dims))
        normalized += bias

    return normalized


def global_response_norm(
    input_: Tensor, gamma: Tensor, beta: Tensor, eps: float = 1e-6
) -> Tensor:
    Gx: Tensor = lucid.linalg.norm(input_, ord=2, axis=(-1, -2), keepdims=True)
    Nx = Gx / (Gx.mean(axis=-1, keepdims=True) + eps)
    return gamma * (input_ * Nx) + beta * input_



from typing import Any, Literal

from lucid._tensor import Tensor
from lucid.nn.init import _dist
from lucid.types import _Scalar

_FanMode = Literal["fan_in", "fan_out"]


def _tensor_check(value: Any) -> None:
    if not isinstance(value, Tensor):
        raise TypeError(f"Expected value to be Tensor got {type(value).__name__}.")


def uniform(tensor: Tensor, a: _Scalar = 0, b: _Scalar = 1) -> None:
    _tensor_check(tensor)
    return _dist.uniform(tensor, a, b)


def normal(tensor: Tensor, mean: _Scalar = 0.0, std: _Scalar = 1.0) -> None:
    _tensor_check(tensor)
    return _dist.normal(tensor, mean, std)


def constant(tensor: Tensor, val: _Scalar) -> None:
    _tensor_check(tensor)
    return _dist.constant(tensor, val)


def xavier_uniform(tensor: Tensor, gain: _Scalar = 1.0) -> None:
    _tensor_check(tensor)
    return _dist.xavier_uniform(tensor, gain)


def xavier_normal(tensor: Tensor, gain: _Scalar = 1.0) -> None:
    _tensor_check(tensor)
    return _dist.xavier_normal(tensor, gain)


def kaiming_uniform(tensor: Tensor, mode: _FanMode = "fan_in") -> None:
    _tensor_check(tensor)
    if mode not in {"fan_in", "fan_out"}:
        raise ValueError("mode must be either 'fan_in' or 'fan_out'.")

    return _dist.kaiming_uniform(tensor, mode)


def kaiming_normal(tensor: Tensor, mode: _FanMode = "fan_in") -> None:
    _tensor_check(tensor)
    if mode not in {"fan_in", "fan_out"}:
        raise ValueError("mode must be either 'fan_in' or 'fan_out'.")

    return _dist.kaiming_normal(tensor, mode)



from dataclasses import dataclass
from typing import Literal, Self, Sequence, TypedDict

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


class _UnionFind:
    def __init__(self, n: int) -> None:
        self.parent = lucid.arange(n, dtype=lucid.Int32)
        self.size = lucid.ones(n, dtype=lucid.Int32)
        self.int_diff = lucid.zeros(n)

    def _p(self, idx: int) -> int:
        return self.parent[idx].item()

    def find(self, x: int) -> int:
        root = x
        while self._p(root) != root:
            root = self._p(root)

        while self._p(x) != x:
            nxt = self._p(x)
            self.parent[x] = root
            x = nxt

        return root

    def union(self, x: int, y: int, weight: float) -> int:
        x_root, y_root = self.find(x), self.find(y)
        if x_root == y_root:
            return x_root

        if self.size[x_root].item() < self.size[y_root].item():
            x_root, y_root = y_root, x_root

        self.parent[y_root] = x_root
        self.size[x_root] = self.size[x_root] + self.size[y_root]

        self.int_diff[x_root] = max(
            self.int_diff[x_root].item(), self.int_diff[y_root].item(), weight
        )
        return x_root

    def component_size(self, x: int) -> int:
        return self.size[self.find(x)].item()


def _compute_edges(
    image: Tensor, connectivity: Literal[4, 8] = 8
) -> tuple[Tensor, Tensor, Tensor]:
    H, W = image.shape[:2]
    idx = lucid.arange(H * W, dtype=lucid.Int32).reshape(H, W)

    def _color_dist(a: Tensor, b: Tensor) -> Tensor:
        diff = a.astype(lucid.Float32) - b.astype(lucid.Float32)
        if diff.ndim == 2:
            return lucid.abs(diff)
        return lucid.sqrt(lucid.sum(diff * diff, axis=-1))

    displacements = [(0, 1), (1, 0)]
    if connectivity == 8:
        displacements += [(1, 1), (1, -1)]

    edges_p, edges_q, edges_w = [], [], []
    for dy, dx in displacements:
        p = idx[max(0, dy) : H - max(0, -dy), max(0, dx) : W - max(0, -dx)].ravel()
        q = idx[max(0, -dy) : H - max(0, dy), max(0, -dx) : W - max(0, dx)].ravel()

        w = _color_dist(
            image[max(0, dy) : H - max(0, -dy), max(0, dx) : W - max(0, -dx)],
            image[max(0, -dy) : H - max(0, dy), max(0, -dx) : W - max(0, dx)],
        ).ravel()

        edges_p.append(p)
        edges_q.append(q)
        edges_w.append(w)

    return (
        lucid.concatenate(edges_p).to(image.device),
        lucid.concatenate(edges_q).to(image.device),
        lucid.concatenate(edges_w).to(image.device),
    )


def felzenszwalb_segmentation(
    image: Tensor, k: float = 500.0, min_size: int = 20, connectivity: Literal[4, 8] = 8
) -> Tensor:
    C, H, W = image.shape
    img_f32 = image.astype(lucid.Float32)
    img_cl = img_f32[0] if C == 1 else img_f32.transpose((1, 2, 0))

    n_px = H * W
    p, q, w = _compute_edges(img_cl, connectivity)
    order = lucid.argsort(w, kind="mergesort")
    p, q, w = p[order], q[order], w[order]

    p_list, q_list, w_list = p.data.tolist(), q.data.tolist(), w.data.tolist()
    uf = _UnionFind(n_px)

    for pi, qi, wi in zip(p_list, q_list, w_list):
        Cp, Cq = uf.find(pi), uf.find(qi)
        if Cp == Cq:
            continue

        thresh = min(
            uf.int_diff[Cp].item() + k / uf.component_size(Cp),
            uf.int_diff[Cq].item() + k / uf.component_size(Cq),
        )
        if wi <= thresh:
            uf.union(Cp, Cq, wi)

    for pi, qi, wi in zip(p_list, q_list, w_list):
        Cp, Cq = uf.find(pi), uf.find(qi)
        if Cp != Cq and (
            uf.component_size(Cp) < min_size or uf.component_size(Cq) < min_size
        ):
            uf.union(Cp, Cq, wi)

    roots = Tensor([uf.find(i) for i in range(n_px)], dtype=lucid.Int32)
    labels = lucid.unique(roots, return_inverse=True)[1]

    return labels.reshape(H, W)


@dataclass
class _Region:
    idx: int
    bbox: tuple[int, int, int, int]
    size: int
    color_hist: Tensor

    def merge(self, other: Self, new_idx: int) -> Self:
        x1 = min(self.bbox[0], other.bbox[0])
        y1 = min(self.bbox[1], other.bbox[1])
        x2 = max(self.bbox[2], other.bbox[2])
        y2 = max(self.bbox[3], other.bbox[3])

        size = self.size + other.size
        color_hist = (
            self.color_hist * self.size + other.color_hist * other.size
        ) / size
        return _Region(new_idx, (x1, y1, x2, y2), size, color_hist)


class SelectiveSearch(nn.Module):
    def __init__(
        self,
        scales: tuple[float, ...] = (50, 100, 150, 300),
        min_size: int = 20,
        connectivity: Literal[4, 8] = 8,
        max_boxes: int = 2000,
        iou_thresh: float = 0.8,
    ) -> None:
        super().__init__()
        self.scales = scales
        self.min_size = min_size
        self.connectivity = connectivity
        self.max_boxes = max_boxes
        self.iou_thresh = iou_thresh

    @staticmethod
    def _color_hist(region_pixels: Tensor, bins: int = 8) -> Tensor:
        hist = lucid.histogramdd(
            region_pixels.reshape(-1, 3), bins, range=((0, 256),) * 3
        )[0].flatten()

        hist_sum = hist.sum()
        return hist / hist_sum if hist_sum.item() else hist

    @lucid.no_grad()
    def forward(self, image: Tensor) -> Tensor:
        if image.ndim != 3:
            raise ValueError("Expecting (C, H, W)")

        _, H, W = image.shape
        rgb = image.transpose((1, 2, 0)).astype(lucid.Int16)
        all_boxes: list[tuple[int, int, int, int]] = []

        for k in self.scales:
            labels = felzenszwalb_segmentation(
                image,
                k=float(k),
                min_size=self.min_size,
                connectivity=self.connectivity,
            )
            n_regions = lucid.max(labels).item() + 1

            regions: dict[int, _Region] = {}
            for rid in range(n_regions):
                mask = labels == rid
                coords = lucid.nonzero(mask.astype(lucid.Int16))
                size = len(coords)
                if size == 0:
                    continue

                ys, xs = coords[:, 0], coords[:, 1]
                bbox = (
                    int(lucid.min(xs).item()),
                    int(lucid.min(ys).item()),
                    int(lucid.max(xs).item()),
                    int(lucid.max(ys).item()),
                )
                color_hist = self._color_hist(rgb[mask.astype(bool)])
                regions[rid] = _Region(rid, bbox, size, color_hist)

            adj: dict[tuple[int, int], float] = {}
            h_a = labels[:, :-1].ravel()
            h_b = labels[:, 1:].ravel()
            v_a = labels[:-1, :].ravel()
            v_b = labels[1:, :].ravel()

            h_neigh = lucid.stack((h_a, h_b), axis=1)
            v_neigh = lucid.stack((v_a, v_b), axis=1)

            def _sim(r1: _Region, r2: _Region) -> float:
                color_sim = lucid.minimum(r1.color_hist, r2.color_hist).sum().item()
                size_sim = 1.0 - (r1.size + r2.size) / float(H * W)

                x1 = min(r1.bbox[0], r2.bbox[0])
                y1 = min(r1.bbox[1], r2.bbox[1])
                x2 = max(r1.bbox[2], r2.bbox[2])
                y2 = max(r1.bbox[3], r2.bbox[3])

                bbox_size = (x2 - x1 + 1) * (y2 - y1 + 1)
                fill_sim = 1.0 - (bbox_size - r1.size - r2.size) / float(H * W)
                return color_sim + size_sim + fill_sim

            for a, b in lucid.concatenate((h_neigh, v_neigh), axis=0):
                ai, bi = int(a.item()), int(b.item())
                if ai == bi:
                    continue
                key = (ai, bi) if ai < bi else (bi, ai)
                adj[key] = _sim(regions[ai], regions[bi])

            for r in regions.values():
                all_boxes.append(r.bbox)

            next_idx = n_regions
            while adj and len(all_boxes) < self.max_boxes:
                (ra, rb), _ = max(adj.items(), key=lambda item: item[1])
                new_region = regions[ra].merge(regions[rb], next_idx)
                next_idx += 1

                del regions[ra]
                del regions[rb]

                neighbors: set[int] = set()
                for i, j in list(adj.keys()):
                    if i in (ra, rb) or j in (ra, rb):
                        adj.pop((i, j))
                        n = j if i in (ra, rb) else i
                        if n not in (ra, rb):
                            neighbors.add(n)

                regions[new_region.idx] = new_region
                all_boxes.append(new_region.bbox)

                for n in neighbors:
                    if n not in regions:
                        continue
                    key = (
                        (n, new_region.idx)
                        if n < new_region.idx
                        else (new_region.idx, n)
                    )
                    adj[key] = _sim(regions[n], new_region)

        unique_boxes: list[tuple[int, int, int, int]] = []
        for b in all_boxes:
            tb = Tensor(b)
            if all(iou(tb, Tensor(ub)) <= self.iou_thresh for ub in unique_boxes):
                unique_boxes.append(b)
            if len(unique_boxes) >= self.max_boxes:
                break

        if unique_boxes:
            return Tensor(unique_boxes, dtype=lucid.Int32)
        return lucid.empty(0, 4, dtype=lucid.Int32)


def iou(boxes_a: Tensor, boxes_b: Tensor) -> Tensor:
    x1a, y1a, x2a, y2a = boxes_a.unbind(axis=1)
    x1b, y1b, x2b, y2b = boxes_b.unbind(axis=1)

    xx1 = lucid.maximum(x1a.unsqueeze(1), x1b.unsqueeze(0))
    yy1 = lucid.maximum(y1a.unsqueeze(1), y1b.unsqueeze(0))
    xx2 = lucid.minimum(x2a.unsqueeze(1), x2b.unsqueeze(0))
    yy2 = lucid.minimum(y2a.unsqueeze(1), y2b.unsqueeze(0))

    w = (xx2 - xx1 + 1).clip(min_value=0)
    h = (yy2 - yy1 + 1).clip(min_value=0)
    inter = w * h

    area_a = (x2a - x1a + 1) * (y2a - y1a + 1)
    area_b = (x2b - x1b + 1) * (y2b - y1b + 1)

    return inter / (area_a.unsqueeze(1) + area_b - inter + 1e-9)


def bbox_to_delta(src: Tensor, target: Tensor, add_one: float = 1.0) -> Tensor:
    sw = src[:, 2] - src[:, 0] + add_one
    sh = src[:, 3] - src[:, 1] + add_one
    sx = src[:, 0] + 0.5 * sw
    sy = src[:, 1] + 0.5 * sh

    tw = target[:, 2] - target[:, 0] + add_one
    th = target[:, 3] - target[:, 1] + add_one
    tx = target[:, 0] + 0.5 * tw
    ty = target[:, 1] + 0.5 * th

    dx = (tx - sx) / sw
    dy = (ty - sy) / sh
    dw = lucid.log(tw / sw)
    dh = lucid.log(th / sh)

    return lucid.stack([dx, dy, dw, dh], axis=1)


def apply_deltas(boxes: Tensor, deltas: Tensor, add_one: float = 1.0) -> Tensor:
    widths = boxes[:, 2] - boxes[:, 0] + add_one
    heights = boxes[:, 3] - boxes[:, 1] + add_one

    ctr_x = boxes[:, 0] + 0.5 * widths
    ctr_y = boxes[:, 1] + 0.5 * heights

    dx, dy, dw, dh = deltas.unbind(axis=-1)
    pred_w = (lucid.exp(dw) * widths).clip(min_value=add_one)
    pred_h = (lucid.exp(dh) * heights).clip(min_value=add_one)

    pred_ctr_x = dx * widths + ctr_x
    pred_ctr_y = dy * heights + ctr_y

    x1 = pred_ctr_x - 0.5 * pred_w
    y1 = pred_ctr_y - 0.5 * pred_h
    x2 = pred_ctr_x + 0.5 * pred_w - add_one
    y2 = pred_ctr_y + 0.5 * pred_h - add_one

    x1, x2 = lucid.minimum(x1, x2), lucid.maximum(x1, x2)
    y1, y2 = lucid.minimum(y1, y2), lucid.maximum(y1, y2)

    return lucid.stack([x1, y1, x2, y2], axis=-1)


def nms(boxes: Tensor, scores: Tensor, iou_thresh: float = 0.3) -> Tensor:
    N = boxes.shape[0]
    if N == 0:
        return lucid.empty(0, device=boxes.device).astype(lucid.Int)

    _, order = scores.sort(descending=True)
    boxes = boxes[order]

    x1, y1, x2, y2 = boxes.unbind(axis=1)
    areas = (x2 - x1 + 1) * (y2 - y1 + 1)

    xx1 = lucid.maximum(x1.unsqueeze(1), x1.unsqueeze(0))
    yy1 = lucid.maximum(y1.unsqueeze(1), y1.unsqueeze(0))

    xx2 = lucid.minimum(x2.unsqueeze(1), x2.unsqueeze(0))
    yy2 = lucid.minimum(y2.unsqueeze(1), y2.unsqueeze(0))

    w = (xx2 - xx1 + 1).clip(min_value=0)
    h = (yy2 - yy1 + 1).clip(min_value=0)
    inter = w * h

    iou = inter / (areas.unsqueeze(1) + areas - inter)

    keep_mask = lucid.ones(N, dtype=bool, device=boxes.device)
    eye = lucid.eye(N, dtype=bool, device=boxes.device)
    for i in range(N):
        if not keep_mask[i]:
            continue
        keep_mask &= (iou[i] <= iou_thresh) | eye[i]

    keep = lucid.nonzero(keep_mask).flatten()
    return order[keep].astype(lucid.Int)


def clip_boxes(boxes: Tensor, image_shape: tuple[int, int]) -> Tensor:
    H, W = image_shape
    x1 = lucid.clip(boxes[:, 0], 0, W - 1)
    y1 = lucid.clip(boxes[:, 1], 0, H - 1)
    x2 = lucid.clip(boxes[:, 2], 0, W - 1)
    y2 = lucid.clip(boxes[:, 3], 0, H - 1)

    return lucid.stack([x1, y1, x2, y2], axis=-1)


class ROIAlign(nn.Module):
    def __init__(self, output_size: tuple[int, int]) -> None:
        super().__init__()
        self.output_size = output_size

    def forward(self, images: Tensor, rois: Tensor, roi_idx: Tensor) -> Tensor:
        C = images.shape[1]
        ph, pw = self.output_size
        device = images.device

        x1, y1, x2, y2 = rois.unbind(axis=1)
        valid = (x2 > x1) & (y2 > y1)
        if lucid.sum(valid) == 0:
            return lucid.empty(0, C, ph, pw, device=device)

        idx = valid.nonzero().squeeze(axis=1)
        rois = rois[idx]
        roi_idx = roi_idx[idx]
        x1, y1, x2, y2 = rois.unbind(axis=1)

        if pw > 1:
            xs = lucid.arange(pw, dtype=lucid.Float32, device=device) / (pw - 1)
        else:
            xs = lucid.full((1,), 0.5, dtype=lucid.Float32, device=device)

        if ph > 1:
            ys = lucid.arange(ph, dtype=lucid.Float32, device=device) / (ph - 1)
        else:
            ys = lucid.full((1,), 0.5, dtype=lucid.Float32, device=device)

        grid_y_base, grid_x_base = lucid.meshgrid(ys, xs, indexing="ij")
        grid_x_base = grid_x_base.unsqueeze(axis=0)
        grid_y_base = grid_y_base.unsqueeze(axis=0)

        x1_ = x1.unsqueeze(axis=-1).unsqueeze(axis=-1)
        y1_ = y1.unsqueeze(axis=-1).unsqueeze(axis=-1)
        w_ = (x2 - x1).unsqueeze(axis=-1).unsqueeze(axis=-1)
        h_ = (y2 - y1).unsqueeze(axis=-1).unsqueeze(axis=-1)

        grid_x = x1_ + w_ * grid_x_base
        grid_y = y1_ + h_ * grid_y_base
        grid = lucid.stack([grid_y * 2 - 1, grid_x * 2 - 1], axis=-1)

        feat_per_roi = images[roi_idx]
        out = F.grid_sample(feat_per_roi, grid, mode="bilinear", align_corners=True)

        return out


class MultiScaleROIAlign(nn.Module):
    def __init__(
        self,
        output_size: tuple[int, int],
        canonical_level: int = 2,
        canocical_scale: int = 224,
    ) -> None:
        super().__init__()
        self.output_size = output_size
        self.canonical_level = canonical_level
        self.canonical_scale = canocical_scale

        self.align = ROIAlign(output_size)

    def forward(self, features: list[Tensor], rois: Tensor, roi_idx: Tensor) -> Tensor:
        device = rois.device

        x1, y1, x2, y2 = rois.unbind(axis=-1)
        widths = x2 - x1
        heights = y2 - y1
        scales = lucid.sqrt(widths * heights)

        target_lvls = lucid.log2(scales / self.canonical_scale + 1e-6)
        target_lvls = target_lvls.astype(lucid.Int32) + self.canonical_level
        target_lvls = lucid.clip(target_lvls, 2, 5)

        pooled = []
        for level in range(2, 6):
            mask = (target_lvls == level).nonzero().squeeze(axis=1)
            if mask.size == 0:
                continue

            feat = features[level - 2]
            rois_l = rois[mask]
            idx_l = roi_idx[mask]
            pooled.append(self.align(feat, rois_l, idx_l))

        if pooled:
            return lucid.concatenate(pooled, axis=0)
        else:
            B, C, ph, pw = features[0].shape[0], features[0].shape[1], *self.output_size
            return lucid.empty(B, C, ph, pw, device=device)


class FPN(nn.Module):
    def __init__(self, in_channels_list: list[int], out_channels: int = 256) -> None:
        super().__init__()
        self.lateral_convs = nn.ModuleList(
            [nn.Conv2d(in_c, out_channels, kernel_size=1) for in_c in in_channels_list]
        )
        self.output_convs = nn.ModuleList(
            [
                nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)
                for _ in in_channels_list
            ]
        )

    def forward(self, features: list[Tensor]) -> list[Tensor]:
        feats = [lateral(f) for f, lateral in zip(features, self.lateral_convs)]
        for i in reversed(range(len(feats) - 1)):
            up = F.interpolate(feats[i + 1], size=feats[i].shape[2:], mode="nearest")
            feats[i] += up

        return [conv(f) for f, conv in zip(feats, self.output_convs)]


class DetectionDict(TypedDict):
    box: tuple[float, float, float, float] | list[float]
    score: float
    class_id: int



from typing import Literal, Callable

import lucid
from lucid._tensor import Tensor

from ._util import _interpolate_nearest


def affine_grid(
    theta: Tensor, size: tuple[int, ...], align_corners: bool = True
) -> Tensor:
    N, _, H, W = size
    device = theta.device

    if align_corners:
        xs = lucid.linspace(-1, 1, W)
        ys = lucid.linspace(-1, 1, H)
    else:
        xs = lucid.linspace(-1 + 1 / W, 1 - 1 / W, W)
        ys = lucid.linspace(-1 + 1 / H, 1 - 1 / H, H)

    x, y = lucid.meshgrid(xs, ys)
    ones = lucid.ones_like(x)

    grid = lucid.stack([x, y, ones], axis=-1)
    grid = grid.reshape(1, H * W, 3).repeat(N, axis=0)
    grid = grid.astype(lucid.Float).to(device).free()

    theta = theta.reshape(N, 2, 3)
    out = grid @ theta.transpose((0, 2, 1))
    out = out.reshape(N, H, W, 2)

    return out


_PaddingType = Literal["zeros", "border"]
_InterpolateType = Literal["bilinear", "nearest"]


def grid_sample(
    input_: Tensor,
    grid: Tensor,
    mode: _InterpolateType = "bilinear",
    padding_mode: _PaddingType = "zeros",
    align_corners: bool = True,
) -> Tensor:
    N, C, H_in, W_in = input_.shape
    N_grid, H_out, W_out, _ = grid.shape
    assert N == N_grid, "Batch size mismatch"

    if align_corners:
        ix = (grid[..., 0] + 1) * (W_in - 1) / 2
        iy = (grid[..., 1] + 1) * (H_in - 1) / 2
    else:
        ix = (grid[..., 0] + 1) * W_in / 2 - 0.5
        iy = (grid[..., 1] + 1) * H_in / 2 - 0.5

    if mode == "nearest":
        ix = ix.round()
        iy = iy.round()

    if padding_mode == "zeros":
        input_ = input_.pad(((0, 0), (0, 0), (1, 1), (1, 1)))
        ix = ix + 1
        iy = iy + 1
    elif padding_mode == "border":
        input_ = _interpolate_nearest(input_, size=(H_in + 2, W_in + 2))
        ix = ix + 1
        iy = iy + 1
    else:
        raise ValueError(f"Unsupported padding_mode: {padding_mode}")

    if mode == "nearest":
        ix = ix.round()
        iy = iy.round()

        ix = lucid.clip(ix, 0, input_.shape[3] - 1).astype(lucid.Int)
        iy = lucid.clip(iy, 0, input_.shape[2] - 1).astype(lucid.Int)

        n_idx = lucid.arange(N)[:, None, None].astype(lucid.Int)
        c_idx = lucid.arange(C)[None, :, None, None].astype(lucid.Int)

        iy = iy[:, None, :, :].repeat(C, axis=1)
        ix = ix[:, None, :, :].repeat(C, axis=1)
        n_idx = n_idx[:, None, :, :].repeat(C, axis=1)

        output = input_[n_idx, c_idx, iy, ix]
        return output

    elif mode == "bilinear":
        x0 = lucid.clip(ix.floor(), 0, input_.shape[3] - 1).astype(lucid.Int)
        x1 = lucid.clip(x0 + 1, 0, input_.shape[3] - 1)

        y0 = lucid.clip(iy.floor(), 0, input_.shape[2] - 1).astype(lucid.Int)
        y1 = lucid.clip(y0 + 1, 0, input_.shape[2] - 1)

        wa = (x1 - ix) * (y1 - iy)
        wb = (x1 - ix) * (iy - y0)
        wc = (ix - x0) * (y1 - iy)
        wd = (ix - x0) * (iy - y0)

        n_idx = lucid.arange(N).reshape(N, 1, 1, 1).astype(lucid.Int)
        c_idx = lucid.arange(C).reshape(1, C, 1, 1).astype(lucid.Int)

        n_idx = n_idx.repeat(C, axis=1).repeat(H_out, axis=2).repeat(W_out, axis=3)
        c_idx = c_idx.repeat(N, axis=0).repeat(H_out, axis=2).repeat(W_out, axis=3)

        def _gather(y: Tensor, x: Tensor) -> Tensor:
            y = y[:, None, :, :].repeat(C, axis=1)
            x = x[:, None, :, :].repeat(C, axis=1)

            return input_[n_idx, c_idx, y, x]

        Ia = _gather(y0, x0)
        Ib = _gather(y1, x0)
        Ic = _gather(y0, x1)
        Id = _gather(y1, x1)

        wa = wa[:, None, :, :].repeat(C, axis=1)
        wb = wb[:, None, :, :].repeat(C, axis=1)
        wc = wc[:, None, :, :].repeat(C, axis=1)
        wd = wd[:, None, :, :].repeat(C, axis=1)

        output = Ia * wa + Ib * wb + Ic * wc + Id * wd
        return output

    else:
        raise ValueError(f"Unsupported interpolation mode: {mode}")



import lucid.nn as nn

from lucid import register_model
from .resnet import ResNet, _Bottleneck


__all__ = [
    "ResNeXt",
    "resnext_50_32x4d",
    "resnext_101_32x4d",
    "resnext_101_32x8d",
    "resnext_101_32x16d",
    "resnext_101_32x32d",
    "resnext_101_64x4d",
]


class ResNeXt(ResNet):
    def __init__(
        self,
        block: nn.Module,
        layers: list[int],
        cardinality: int,
        base_width: int,
        num_classes: int = 1000,
    ) -> None:
        block_args = {"cardinality": cardinality, "base_width": base_width}
        super().__init__(block, layers, num_classes, block_args=block_args)


@register_model
def resnext_50_32x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 6, 3]
    return ResNeXt(_Bottleneck, layers, 32, 4, num_classes, **kwargs)


@register_model
def resnext_101_32x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    return ResNeXt(_Bottleneck, layers, 32, 4, num_classes, **kwargs)


@register_model
def resnext_101_32x8d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    return ResNeXt(_Bottleneck, layers, 32, 8, num_classes, **kwargs)


@register_model
def resnext_101_32x16d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    return ResNeXt(_Bottleneck, layers, 32, 16, num_classes, **kwargs)


@register_model
def resnext_101_32x32d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    return ResNeXt(_Bottleneck, layers, 32, 32, num_classes, **kwargs)


@register_model
def resnext_101_64x4d(num_classes: int = 1000, **kwargs) -> ResNeXt:
    layers = [3, 4, 23, 3]
    return ResNeXt(_Bottleneck, layers, 64, 4, num_classes, **kwargs)



import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = ["VGGNet", "vggnet_11", "vggnet_13", "vggnet_16", "vggnet_19"]


class VGGNet(nn.Module):
    def __init__(self, conv_config: list[int | str], num_classes: int = 1000) -> None:
        super().__init__()
        self.conv = self._make_layers(conv_config)
        self.avgpool = nn.AdaptiveAvgPool2d(output_size=(7, 7))
        self.fc = nn.Sequential(
            nn.Linear(512 * 7 * 7, 4096),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(4096, num_classes),
        )

    def _make_layers(self, config: list[int | str]) -> nn.Sequential:
        layers = []
        in_channels = 3
        for layer in config:
            if layer == "M":
                layers.append(nn.MaxPool2d(kernel_size=2, stride=2))
            else:
                layers.append(nn.Conv2d(in_channels, layer, kernel_size=3, padding=1))
                layers.append(nn.ReLU())

                in_channels = layer

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.conv(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)
        return x


@register_model
def vggnet_11(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, "M"]
    config.extend([128, "M"])
    config.extend([256, 256, "M"])
    config.extend([512, 512, "M", 512, 512, "M"])

    return VGGNet(config, num_classes, **kwargs)


@register_model
def vggnet_13(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, "M"])
    config.extend([512, 512, "M", 512, 512, "M"])

    return VGGNet(config, num_classes, **kwargs)


@register_model
def vggnet_16(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, 256, "M"])
    config.extend([512, 512, 512, "M", 512, 512, 512, "M"])

    return VGGNet(config, num_classes, **kwargs)


@register_model
def vggnet_19(num_classes: int = 1000, **kwargs) -> VGGNet:
    config = [64, 64, "M"]
    config.extend([128, 128, "M"])
    config.extend([256, 256, 256, 256, "M"])
    config.extend([512, 512, 512, 512, "M", 512, 512, 512, 512, "M"])

    return VGGNet(config, num_classes, **kwargs)



from typing import Callable, Literal

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["VAE"]


class VAE(nn.Module):
    def __init__(
        self,
        encoders: list[nn.Module],
        decoders: list[nn.Module],
        priors: list[nn.Module] | None,
        reconstruction_loss: Literal["mse", "bce"] = "mse",
        kl_weight: float = 1.0,
        beta_schedule: Callable[[int], float] | None = None,
        hierarchical_kl: bool = True,
        depth: int | None = None,
    ) -> None:
        super().__init__()
        self.depth = depth or len(encoders)
        assert (
            self.depth == len(encoders) == len(decoders)
        ), "Mismatch in component lengths"

        self.encoders = nn.ModuleList(encoders)
        self.decoders = nn.ModuleList(decoders)
        self.priors = nn.ModuleList(priors) if priors is not None else None

        self.reconstruction_loss = reconstruction_loss
        self.base_kl_weight = kl_weight
        self.beta_schedule = beta_schedule
        self.hierarchical_kl = hierarchical_kl

        self._training_step = 0

    def reparameterize(self, mu: Tensor, logvar: Tensor) -> Tensor:
        std = lucid.exp(0.5 * logvar)
        eps = lucid.random.randn(std.shape, device=std.device)
        return mu + eps * std

    def encode(self, x: Tensor) -> tuple[list[Tensor], ...]:
        mus: list[Tensor] = []
        logvars: list[Tensor] = []
        zs: list[Tensor] = []
        h = x
        for encoder in self.encoders:
            h = encoder(h)
            mu, logvar = lucid.split(h, 2, axis=1)
            z = self.reparameterize(mu, logvar)

            mus.append(mu)
            logvars.append(logvar)
            zs.append(z)
            h = z

        return zs, mus, logvars

    def decode(self, zs: list[Tensor]) -> Tensor:
        h = zs[-1]
        for decoder in reversed(self.decoders):
            h = decoder(h)
        return h

    def forward(
        self, x: Tensor
    ) -> tuple[Tensor, list[Tensor], list[Tensor], list[Tensor]]:
        zs, mus, logvars = self.encode(x)
        recon = self.decode(zs)
        return recon, mus, logvars, zs

    def compute_kl(
        self, mu_q: Tensor, logvar_q: Tensor, mu_p: Tensor, logvar_p: Tensor
    ) -> Tensor:
        total_kl = 0.5 * (
            logvar_p
            - logvar_q
            + (lucid.exp(logvar_q) + (mu_q - mu_p) ** 2) / lucid.exp(logvar_p)
            - 1
        )
        return total_kl.sum(axis=1).mean()

    def current_beta(self) -> float:
        if self.beta_schedule is not None:
            return self.beta_schedule(self._training_step)
        return self.base_kl_weight

    def get_loss(
        self,
        x: Tensor,
        recon: Tensor,
        mus: list[Tensor],
        logvars: list[Tensor],
        zs: list[Tensor],
    ) -> tuple[Tensor, Tensor, Tensor]:
        beta = self.current_beta()
        self._training_step += 1

        if self.reconstruction_loss == "mse":
            recon_loss = F.mse_loss(recon, x)
        elif self.reconstruction_loss == "bce":
            recon = recon.clip(1e-7, 1 - 1e-7)
            recon_loss = F.binary_cross_entropy(recon, x)
        else:
            raise ValueError(
                f"Unsupported reconstruction loss: {self.reconstruction_loss}"
            )

        kl = Tensor(0.0)
        for l in range(self.depth):
            mu_q, logvar_q = mus[l], logvars[l]
            if l == self.depth - 1:
                mu_p = lucid.zeros_like(mu_q)
                logvar_p = lucid.zeros_like(logvar_q)
            elif self.priors is not None:
                prior_out = self.priors[l](zs[l + 1])
                mu_p, logvar_p = lucid.split(prior_out, 2, axis=1)
            else:
                mu_p = lucid.zeros_like(mu_q)
                logvar_p = lucid.zeros_like(logvar_q)

            scale = beta / self.depth if self.hierarchical_kl else beta
            kl += scale * self.compute_kl(mu_q, logvar_q, mu_p, logvar_p)

        return recon_loss + kl, recon_loss, kl



from lucid._tensor import Tensor
from lucid.types import _ArrayOrScalar, _DeviceType


__all__ = ["Parameter", "Buffer"]


class Parameter(Tensor):
    def __init__(
        self,
        data: Tensor | _ArrayOrScalar,
        dtype: type | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(
            data, requires_grad=True, keep_grad=True, dtype=dtype, device=device
        )


class Buffer(Tensor):
    def __init__(
        self,
        data: Tensor | _ArrayOrScalar,
        dtype: type | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        if isinstance(data, Tensor):
            data = data.data
        super().__init__(
            data, requires_grad=False, keep_grad=False, dtype=dtype, device=device
        )



from typing import Literal, Sequence, overload

from lucid._tensor import Tensor
from lucid.types import _ShapeLike, _ArrayLikeInt, _Scalar

from lucid._util import func


# fmt: off
__all__ = [
    "reshape", "squeeze", "unsqueeze", "expand_dims", "ravel", "stack", "hstack",
    "vstack", "concatenate", "pad", "repeat", "tile", "flatten", "meshgrid", 
    "split", "tril", "triu", "broadcast_to", "chunk", "masked_fill", "roll", 
    "unbind", "sort", "nonzero", "unique", "topk", "argsort", "histogramdd", 
    "histogram", "histogram2d", "where", "nonzero", "argmin", "argmax", 
    "diagonal",
]
# fmt: on


def reshape(a: Tensor, /, shape: _ShapeLike) -> Tensor:
    return func.reshape(shape)(a)


@overload
def reshape_immediate(a: Tensor, /, shape: _ShapeLike) -> Tensor: ...


def _reshape_immediate(a: Tensor, /, *shape: int | _ShapeLike) -> Tensor:
    if isinstance(shape[0], (tuple, list)):
        shape = shape[0]
    return func._reshape_immediate(shape)(a)


def squeeze(a: Tensor, /, axis: _ShapeLike | None = None) -> Tensor:
    return func.squeeze(axis)(a)


def unsqueeze(a: Tensor, /, axis: _ShapeLike) -> Tensor:
    return func.unsqueeze(axis)(a)


def expand_dims(a: Tensor, /, axis: _ShapeLike) -> Tensor:
    return func.expand_dims(axis)(a)


def ravel(a: Tensor, /) -> Tensor:
    return func.ravel()(a)


def stack(arr: tuple[Tensor, ...], /, axis: int = 0) -> Tensor:
    return func.stack(axis)(*arr)


def hstack(arr: tuple[Tensor, ...], /) -> Tensor:
    return func.hstack()(*arr)


def vstack(arr: tuple[Tensor, ...], /) -> Tensor:
    return func.vstack()(*arr)


def concatenate(arr: tuple[Tensor, ...], /, axis: int = 0) -> Tensor:
    return func.concatenate(axis)(*arr)


def pad(a: Tensor, /, pad_width: _ArrayLikeInt) -> Tensor:
    return func.pad(pad_width, ndim=a.ndim)(a)


def repeat(
    a: Tensor, /, repeats: int | Sequence[int], axis: int | None = None
) -> Tensor:
    return func.repeat(repeats, axis)(a)


def tile(a: Tensor, /, reps: int | Sequence[int]) -> Tensor:
    return func.tile(reps)(a)


def flatten(a: Tensor, /, start_axis: int = 0, end_axis: int = -1) -> Tensor:
    return func.flatten(start_axis, end_axis)(a)


def meshgrid(
    a: Tensor, b: Tensor, /, indexing: Literal["xy", "ij"] = "ij"
) -> tuple[Tensor, Tensor]:
    return func.meshgrid(indexing)(a, b)


def split(
    a: Tensor, /, size_or_sections: int | list[int] | tuple[int], axis: int = 0
) -> tuple[Tensor, ...]:
    return func.split(size_or_sections, axis)(a)


def tril(a: Tensor, /, diagonal: int = 0) -> Tensor:
    return func.tril(diagonal)(a)


def triu(a: Tensor, /, diagonal: int = 0) -> Tensor:
    return func.triu(diagonal)(a)


def broadcast_to(a: Tensor, /, shape: _ShapeLike) -> Tensor:
    return func.broadcast_to(shape)(a)


def chunk(a: Tensor, /, chunks: int, axis: int = 0) -> tuple[Tensor, ...]:
    return func.chunk(chunks, axis)(a)


def masked_fill(a: Tensor, /, mask: Tensor, value: _Scalar) -> Tensor:
    return func.masked_fill(mask, value)(a)


def roll(
    a: Tensor,
    /,
    shifts: int | tuple[int, ...],
    axis: int | tuple[int, ...] | None = None,
) -> Tensor:
    return func.roll(shifts, axis)(a)


def unbind(a: Tensor, /, axis: int = 0) -> tuple[Tensor, ...]:
    return func.unbind(axis)(a)


_SortKind = Literal["quicksort", "mergesort", "heapsort", "stable"]


def sort(
    a: Tensor,
    /,
    axis: int = -1,
    descending: bool = False,
    kind: _SortKind | None = None,
    stable: bool = False,
) -> tuple[Tensor, Tensor]:
    return func.sort(axis, descending, kind, stable)(a)


def argsort(
    a: Tensor,
    /,
    axis: int = -1,
    descending: bool = False,
    kind: _SortKind | None = None,
    stable: bool = False,
) -> Tensor:
    return func.argsort(axis, descending, kind, stable)(a)


def argmin(a: Tensor, axis: int | None = None, keepdims: bool = False) -> Tensor:
    return func.argmin(axis, keepdims)(a)


def argmax(a: Tensor, axis: int | None = None, keepdims: bool = False) -> Tensor:
    return func.argmax(axis, keepdims)(a)


def nonzero(a: Tensor, /) -> Tensor:
    return func.nonzero()(a)


def unique(
    a: Tensor,
    /,
    sorted: bool = True,
    axis: int | None = None,
    return_inverse: bool = False,
) -> Tensor | tuple[Tensor, Tensor]:
    op = func.unique(sorted, axis)
    ret = op(a)
    if return_inverse:
        assert (
            op.inverse_ is not None
        ), "inverse_ was not computed; check axis/sorted constraints"
        return ret, op.inverse_
    else:
        return ret


def topk(
    a: Tensor, /, k: int, axis: int = -1, largest: bool = True, sorted: bool = True
) -> tuple[Tensor, Tensor]:
    if k > a.shape[axis]:
        raise ValueError(
            f"k={k} is greater than dimension size {a.shape[axis]} along axis {axis}."
        )
    return func.topk(k, axis, largest, sorted)(a)


def histogramdd(
    a: Tensor,
    /,
    bins: int | list[int],
    range: list[tuple[float, float]],
    density: bool = False,
) -> tuple[Tensor, Tensor]:
    if isinstance(bins, int):
        bins = [bins] * a.shape[1]
    return func.histogramdd(bins, range, density)(a)


def histogram(
    a: Tensor,
    /,
    bins: int = 10,
    range: tuple[float, float] | None = None,
    density: bool = False,
) -> tuple[Tensor, Tensor]:
    if a.ndim != 1:
        raise ValueError("histogram() expects a 1D tensor input.")
    a = a.reshape(-1, 1)
    range = [range or (float(a.data.min().item()), float(a.data.max().item()))]
    return func.histogramdd([bins], range, density)(a)


def histogram2d(
    a: Tensor,
    b: Tensor,
    /,
    bins: list[int, int] = [10, 10],
    range: list[tuple[float, float]] | None = None,
    density: bool = False,
) -> tuple[Tensor, Tensor]:
    if a.shape != b.shape:
        raise ValueError("a and b must have the same shape.")
    ab = stack((a, b), axis=1)
    if range is None:
        range = [
            (float(a.data.min().item()), float(a.data.max().item())),
            (float(b.data.min().item()), float(b.data.max().item())),
        ]
    return func.histogramdd(bins, range, density)(ab)


def where(condition: Tensor, a: Tensor, b: Tensor, /) -> Tensor:
    return func.where()(condition, a, b)


def diagonal(a: Tensor, /, offset: int = 0, axis1: int = 0, axis2: int = 1) -> Tensor:
    return func.diagonal(offset, axis1, axis2)(a)


Tensor.reshape = _reshape_immediate
Tensor.squeeze = squeeze
Tensor.unsqueeze = unsqueeze
Tensor.ravel = ravel
Tensor.pad = pad
Tensor.repeat = repeat
Tensor.tile = tile
Tensor.flatten = flatten
Tensor.split = split
Tensor.tril = tril
Tensor.triu = triu
Tensor.broadcast_to = broadcast_to
Tensor.chunk = chunk
Tensor.masked_fill = masked_fill
Tensor.roll = roll
Tensor.unbind = unbind
Tensor.sort = sort
Tensor.unique = unique
Tensor.nonzero = nonzero
Tensor.diagonal = diagonal



import numpy as np

import lucid
from lucid._tensor import Tensor
from lucid.types import (
    _ShapeLike,
    _Scalar,
    _ArrayOrScalar,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)

from lucid._backend.metal import mx


def seed(seed: int) -> None:
    np.random.seed(seed)
    mx.random.seed(seed)


def rand(
    shape: _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.random.rand(*shape), requires_grad, keep_grad, device=device)


def randint(
    low: int,
    high: int | None,
    size: int | _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(
        np.random.randint(low, high, size),
        requires_grad,
        keep_grad,
        lucid.Int64,
        device,
    )


def randn(
    shape: _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(np.random.randn(*shape), requires_grad, keep_grad, device=device)


def uniform(
    low: _Scalar,
    high: _Scalar,
    size: int | _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return Tensor(
        np.random.uniform(low, high, size), requires_grad, keep_grad, device=device
    )


def bernoulli(
    probs: _ArrayOrScalar | Tensor,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    if isinstance(probs, Tensor):
        probs_data = probs.data
    else:
        probs_data = np.array(probs)

    if np.any(probs_data < 0) or np.any(probs_data > 1):
        raise ValueError("probs must be in the range [0, 1].")

    return Tensor(
        (np.random.rand(*probs_data.shape) < probs_data).astype(int),
        requires_grad,
        keep_grad,
        device=device,
    )


def permutation(
    n: int,
    dtype: _BuiltinNumeric | Numeric,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    data = np.random.permutation(n)
    return Tensor(data, requires_grad, keep_grad, dtype, device)



from typing import Any, ClassVar, Type, Literal
import math

import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "ResNet",
    "resnet_18",
    "resnet_34",
    "resnet_50",
    "resnet_101",
    "resnet_152",
    "resnet_200",
    "resnet_269",
    "resnet_1001",
    "wide_resnet_50",
    "wide_resnet_101",
]


class ResNet(nn.Module):
    def __init__(
        self,
        block: nn.Module,
        layers: list[int],
        num_classes: int = 1000,
        in_channels: int = 3,
        stem_width: int = 64,
        stem_type: Literal["deep"] | None = None,
        avg_down: bool = False,
        channels: tuple[int] = (64, 128, 256, 512),
        block_args: dict[str, Any] = {},
    ) -> None:
        super().__init__()
        deep_stem = stem_type == "deep"
        self.in_channels = stem_width * 2 if deep_stem else 64
        self.avg_down = avg_down

        if deep_stem:
            self.stem = nn.Sequential(
                nn.Conv2d(in_channels, stem_width, 3, stride=2, padding=1, bias=False),
                nn.BatchNorm2d(stem_width),
                nn.ReLU(),
                nn.Conv2d(stem_width, stem_width, 3, stride=1, padding=1, bias=False),
                nn.BatchNorm2d(stem_width),
                nn.ReLU(),
                nn.Conv2d(stem_width, self.in_channels, 3, padding=1, bias=False),
            )
        else:
            self.stem = nn.Sequential(
                nn.Conv2d(
                    in_channels, self.in_channels, 7, stride=2, padding=3, bias=False
                ),
                nn.BatchNorm2d(self.in_channels),
                nn.ReLU(),
            )

        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(
            block, channels[0], layers[0], stride=1, block_args=block_args
        )
        self.layer2 = self._make_layer(
            block, channels[1], layers[1], stride=2, block_args=block_args
        )
        self.layer3 = self._make_layer(
            block, channels[2], layers[2], stride=2, block_args=block_args
        )
        self.layer4 = self._make_layer(
            block, channels[3], layers[3], stride=2, block_args=block_args
        )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(
        self,
        block: Type[nn.Module],
        out_channels: int,
        blocks: int,
        stride: int = 1,
        block_args: dict[str, Any] = {},
    ) -> nn.Sequential:
        downsample = None
        if stride != 1 or self.in_channels != out_channels * block.expansion:
            if self.avg_down:
                downsample = nn.Sequential(
                    nn.AvgPool2d(kernel_size=3, stride=stride, padding=1),
                    nn.Conv2d(
                        self.in_channels,
                        out_channels * block.expansion,
                        kernel_size=1,
                        stride=1,
                        bias=False,
                    ),
                    nn.BatchNorm2d(out_channels * block.expansion),
                )
            else:
                downsample = nn.Sequential(
                    nn.Conv2d(
                        self.in_channels,
                        out_channels * block.expansion,
                        kernel_size=1,
                        stride=stride,
                        bias=False,
                    ),
                    nn.BatchNorm2d(out_channels * block.expansion),
                )

        layers = [
            block(self.in_channels, out_channels, stride, downsample, **block_args)
        ]
        self.in_channels = out_channels * block.expansion

        for _ in range(1, blocks):
            layers.append(block(self.in_channels, out_channels, stride=1, **block_args))

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.maxpool(x)

        for layer in [self.layer1, self.layer2, self.layer3, self.layer4]:
            x = layer(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.fc(x)

        return x


class _BasicBlock(nn.Module):
    expansion: ClassVar[int] = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
    ) -> None:
        super().__init__()
        basic_conv_args = dict(kernel_size=3, padding=1, bias=False)

        self.conv1 = nn.Conv2d(
            in_channels, out_channels, stride=stride, **basic_conv_args
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu1 = nn.ReLU()

        self.conv2 = nn.Conv2d(out_channels, out_channels, stride=1, **basic_conv_args)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.relu2 = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.relu1(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu2(out)

        return out


class _Bottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        cardinality: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        se: bool = False,
        se_args: dict = {},
    ) -> None:
        super().__init__()
        width = int(math.floor(out_channels * (base_width / 64)) * cardinality)

        self.conv1 = nn.ConvBNReLU2d(
            in_channels, width, kernel_size=1, stride=1, conv_bias=False
        )
        self.conv2 = nn.ConvBNReLU2d(
            width,
            width,
            kernel_size=3,
            stride=stride,
            padding=1,
            dilation=dilation,
            groups=cardinality,
            conv_bias=False,
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(
                width,
                out_channels * self.expansion,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

        self.se = nn.SEModule(out_channels * self.expansion, **se_args) if se else None
        self.relu = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)

        if self.se is not None:
            out = self.se(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class _PreActBottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
    ) -> None:
        super().__init__()

        self.bn1 = nn.BatchNorm2d(in_channels)
        self.relu1 = nn.ReLU()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        self.bn2 = nn.BatchNorm2d(out_channels)
        self.relu2 = nn.ReLU()
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        )

        self.bn3 = nn.BatchNorm2d(out_channels)
        self.relu3 = nn.ReLU()
        self.conv3 = nn.Conv2d(
            out_channels, out_channels * self.expansion, kernel_size=1, bias=False
        )

        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.relu1(self.bn1(x))
        if self.downsample is not None:
            identity = self.downsample(out)

        out = self.conv1(out)
        out = self.conv2(self.relu2(self.bn2(out)))
        out = self.conv3(self.relu3(self.bn3(out)))

        out += identity
        return out


@register_model
def resnet_18(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [2, 2, 2, 2]
    return ResNet(_BasicBlock, layers, num_classes, **kwargs)


@register_model
def resnet_34(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 4, 6, 3]
    return ResNet(_BasicBlock, layers, num_classes, **kwargs)


@register_model
def resnet_50(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 4, 6, 3]
    return ResNet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def resnet_101(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 4, 23, 3]
    return ResNet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def resnet_152(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 8, 36, 3]
    return ResNet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def resnet_200(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 24, 36, 3]
    return ResNet(_PreActBottleneck, layers, num_classes, **kwargs)


@register_model
def resnet_269(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 30, 48, 8]
    return ResNet(_PreActBottleneck, layers, num_classes, **kwargs)


@register_model
def resnet_1001(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 94, 94, 3]
    return ResNet(_PreActBottleneck, layers, num_classes, **kwargs)


@register_model
def wide_resnet_50(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 4, 6, 3]
    block_args = {"base_width": 128}
    return ResNet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)


@register_model
def wide_resnet_101(num_classes: int = 1000, **kwargs) -> ResNet:
    layers = [3, 4, 23, 3]
    block_args = {"base_width": 128}
    return ResNet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)



from typing import Any, Callable, Dict, Sequence, Literal, TypeAlias
import re

# NOTE: This module retains module independency.
import numpy as np
import mlx.core as mx


_DeviceType = Literal["cpu", "gpu"]

_Scalar = int | float | complex
_NumPyArray: TypeAlias = np.ndarray
_MLXArray: TypeAlias = mx.array

_Gradient = _NumPyArray | _MLXArray | None

_ArrayOrScalar = _Scalar | list[_Scalar] | _NumPyArray | _MLXArray

_BuiltinNumeric = type[bool | int | float | complex]

_ShapeLike = list[int] | tuple[int]

_ArrayLike = list | _NumPyArray | _MLXArray
_ArrayLikeInt = int | Sequence[int | tuple[int, int]]

_OptimClosure = Callable[[], Any]

_EinopsPattern = str


class Numeric:
    def __init__(
        self, base_dtype: type[int | float | complex], bits: int | None
    ) -> None:
        self.base_dtype = base_dtype
        self.base_str = base_dtype.__name__
        self.bits = bits

        self._np_dtype: type | None = None
        self._mlx_dtype: type | None = None

        if bits is not None:
            self._np_dtype = getattr(np, self.base_str + str(bits))
            bits_mlx = bits
            if (
                mx.default_device().type is mx.DeviceType.gpu
                and self.base_dtype is float
                and bits == 64
            ):
                bits_mlx = 32
            self._mlx_dtype = getattr(mx, self.base_str + str(bits_mlx))

    @property
    def cpu(self) -> type | None:
        return self._np_dtype

    @property
    def gpu(self) -> type | None:
        return self._mlx_dtype

    @property
    def is_bit_free(self) -> bool:
        return self.bits is None

    def parse(self, device: _DeviceType) -> type | None:
        if device == "cpu":
            return self.cpu
        else:
            return self.gpu

    def auto_parse(self, data_dtype: type, device: _DeviceType) -> type | None:
        bits = self._dtype_bits(data_dtype)
        new_dtype = self.base_dtype.__name__ + str(bits)

        return getattr(np if device == "cpu" else mx, new_dtype, None)

    def _dtype_bits(self, dtype: type) -> int:
        if isinstance(dtype, (np.dtype, type)) and hasattr(dtype, "itemsize"):
            return np.dtype(dtype).itemsize * 8

        if isinstance(dtype, (mx.Dtype, type)):
            return dtype.size * 8

        if isinstance(dtype, str):
            try:
                return np.dtype(dtype).itemsize * 8
            except TypeError:
                return dtype.size * 8

        raise TypeError(f"Unsupported dtype: {dtype}")

    def __str__(self) -> str:
        return self.base_dtype.__name__ + str(self.bits)

    def __repr__(self) -> str:
        return (
            f"(base_dtype={self.base_dtype.__name__}, bits={self.bits},"
            + f" _np_dtype={self.cpu}, _mlx_dtype={self.gpu})",
        )[0]


Int = Numeric(int, None)
Int8 = Numeric(int, bits=8)
Int16 = Numeric(int, bits=16)
Int32 = Numeric(int, bits=32)
Int64 = Numeric(int, bits=64)

Float = Numeric(float, None)
Float16 = Numeric(float, bits=16)
Float32 = Numeric(float, bits=32)
Float64 = Numeric(float, bits=64)

Complex = Numeric(complex, None)
Complex64 = Numeric(complex, bits=64)

numeric_dict = {
    "int": {"8": Int8, "16": Int16, "32": Int32, "64": Int64},
    "float": {"16": Float16, "32": Float32, "64": Float64},
    "complex": {"64": Complex64},
}


def to_numeric_type(data_dtype: type) -> Numeric:
    str_dtype = str(data_dtype).split(".")[-1]

    name = re.findall(r"[a-z]+", str_dtype)[0]
    bits = re.findall(r"\d+", str_dtype)[0]

    return numeric_dict[name][bits]



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["Embedding"]


class Embedding(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: int | None = None,
        max_norm: float | None = None,
        norm_type: float = 2.0,
        _weight: Tensor | None = None,
    ) -> None:
        super().__init__()
        self.num_embeddings = num_embeddings
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        self.max_norm = max_norm
        self.norm_type = norm_type

        if _weight is None:
            self.weight = nn.Parameter(
                lucid.random.uniform(-0.1, 0.1, (num_embeddings, embedding_dim))
            )
        else:
            self.weight = nn.Parameter(_weight)

    def forward(self, input_: Tensor) -> Tensor:
        return F.embedding(
            input_, self.weight, self.padding_idx, self.max_norm, self.norm_type
        )

    def reset_parameters(self) -> None:
        self.weight.data = lucid.random.uniform(-0.1, 0.1, self.weight.shape)



from typing import overload

import lucid
from lucid.random import _func
from lucid._tensor import Tensor
from lucid.types import (
    _ShapeLike,
    _Scalar,
    _ArrayOrScalar,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)


# fmt: off
__all__ = [
    "seed", "rand", "randint", "randn", "uniform", "bernoulli", "permutation"
]
# fmt: on


def seed(seed: int) -> None:
    return _func.seed(seed)


@overload
def rand(
    *shape: int,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu"
) -> Tensor: ...


@overload
def rand(
    shape: _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor: ...


def rand(
    *args: int,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu"
) -> Tensor:
    shape = lucid._get_overloaded_shape(args)
    return _func.rand(shape, requires_grad, keep_grad, device)


def randint(
    low: int,
    high: int | None,
    size: int | _ShapeLike = 1,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return _func.randint(low, high, size, requires_grad, keep_grad, device)


@overload
def randn(
    *shape: int,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu"
) -> Tensor: ...


@overload
def randn(
    shape: _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor: ...


def randn(
    *args: int | _ShapeLike,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu"
) -> Tensor:
    shape = lucid._get_overloaded_shape(args)
    return _func.randn(shape, requires_grad, keep_grad, device)


def uniform(
    low: _Scalar = 0,
    high: _Scalar = 1,
    size: int | _ShapeLike = 1,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return _func.uniform(low, high, size, requires_grad, keep_grad, device)


def bernoulli(
    probs: _ArrayOrScalar | Tensor,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return _func.bernoulli(probs, requires_grad, keep_grad, device)


def permutation(
    n: int,
    dtype: _BuiltinNumeric | Numeric = int,
    requires_grad: bool = False,
    keep_grad: bool = False,
    device: _DeviceType = "cpu",
) -> Tensor:
    return _func.permutation(n, dtype, requires_grad, keep_grad, device)
